"""
This module contains :class:`CoffeeAgent`, a toy agent that is used for testing.
Testing agents may require recording requests and responses in hand-crafted
scenarios, and while Example Agent seems to fit the purpose, we avoid it for two
reasons:

* Some scenarios may be too peculiar to be included there
* A breaking change in the Agent would require generating all the service
  responses from scratch, as tests must match responses with the CoffeeAgent
  class. Example Agent is there for demonstration purposes, and it should be
  possible to update its definition at any time to better express the potential
  of *Intents*.
"""
from typing import Union, List
from dataclasses import dataclass
from unittest.mock import patch

from intents import Agent, Intent, Entity, follow, LanguageCode
from intents.language.intent_language import IntentLanguageData, ExampleUtterance, IntentResponseGroup, TextIntentResponse
from intents.language.entity_language import EntityEntry

def mock_language_data(
    intent_cls: type,
    utterances: Union[str, List[str]],
    responses: Union[str, List[str]]="Any response",
    language: LanguageCode = LanguageCode.ENGLISH
):
    if isinstance(utterances, str):
        utterances = [utterances]
    if isinstance(responses, str):
        responses = [responses]
    return {
        language: IntentLanguageData(
            example_utterances=[ExampleUtterance(u, intent_cls) for u in utterances],
            responses={
                IntentResponseGroup.DEFAULT: [TextIntentResponse(choices=responses)]
            }
        )
    }

class CoffeeRoast(Entity):
    __entity_language_data__ = {
        LanguageCode.ENGLISH: [
            EntityEntry("light", []),
            EntityEntry("medium", []),
            EntityEntry("dark", []),
        ]
    }

@dataclass
class AskCoffee(Intent):
    """I'd like a coffee"""
    name = "AskCoffee"
    roast: CoffeeRoast = "medium"
AskCoffee.__intent_language_data__ = mock_language_data(
    AskCoffee,
    ["I'd like a coffee", "I'd like a $roast{medium} roast coffee"],
    ["$roast roast coffee, good choice!", "Alright, $roast roasted coffee for you"]
)

@dataclass
class AskEspresso(AskCoffee):
    """I'd like an espresso."""
    name = "AskEspresso"
AskEspresso.__intent_language_data__ = mock_language_data(
    AskEspresso,
    ["I'd like an espresso", "I'd like a $roast{medium} roast espresso"],
    ["$roast roast espresso, good choice!", "Alright, $roast roasted espresso for you"]
)
    
@dataclass
class AddMilk(Intent):
    """With milk please"""    
    parent_ask_coffee: AskCoffee = follow()
    name = "AddMilk"
AddMilk.__intent_language_data__ = mock_language_data(AddMilk, "With milk please")

@dataclass
class AddSkimmedMilk(AddMilk):
    """With skimmed milk please"""    
    parent_ask_coffee: AskCoffee = follow()
    name = "AddSkimmedMilk"
AddSkimmedMilk.__intent_language_data__ = mock_language_data(AddMilk, "With skimmed milk please")

@dataclass
class AndNoFoam(Intent):
    """And no foam"""    
    parent_add_milk: AddMilk = follow()
    name = "AndNoFoam"
AndNoFoam.__intent_language_data__ = mock_language_data(AddMilk, "And no foam")

class CoffeeAgent(Agent):
    languages = ['en']

CoffeeAgent.register(AskCoffee)
CoffeeAgent.register(AskEspresso)
CoffeeAgent.register(AddMilk)
CoffeeAgent.register(AddSkimmedMilk)
CoffeeAgent.register(AndNoFoam)



"""
An **intent** is a categorical representation of the User intention in a single
conversation turn. For instance, utterances like "I want a pizza", "I'd like to
order a pizza" and such, could be mapped to a single `OrderPizza` intent.

Your agent will typically define a number of *intents*, representing all the
types of messages the Agent can understand and answer to. This is done by
defining :class:`Intent` sub-classes and their language resources (see
:mod:`intents.language`), and registering them to an :class:`intents.Agent`
class with :meth:`intents.Agent.register`.
"""
import inspect
import logging
import dataclasses
from dataclasses import dataclass
from typing import List, Dict, Union, Any, Type, _GenericAlias

# pylint: disable=unused-import # needed for docs
import intents
from intents import LanguageCode
from intents.model import entity, names
from intents.helpers.data_classes import is_dataclass_strict

logger = logging.getLogger(__name__)

#
# Intent
#

@dataclass
class IntentParameterMetadata:
    """
    Model metadata of a single Intent parameter. `IntentParameterMetadata`
    objects are built internally by :attr:`Intent.parameter_schema` based on the
    Intent dataclass fields.

    >>> OrderPizzaIntent.parameter_schema
    {
        'pizza_type': IntentParameterMetadata(name='pizza_type', entity_cls=<...PizzaType'>, is_list=False, required=True, default=None),
        'amount': IntentParameterMetadata(name='amount', entity_cls=<...Sys.Integer'>, is_list=False, required=False, default=1)
    }

    Args:
        name: Parameter name
        entity_cls: Parameter type
        is_list: Parameter will match multiple values in the User utterance
        required: If True, user will be prompted for parameter value when that
            can't be tagged in his utterance
        default: If set, this value will be used when parameter value can't be
            tagged in the User's utterance
    """
    name: str
    entity_cls: entity.EntityType
    is_list: bool
    required: bool
    default: Any

@dataclass
class FulfillmentContext:
    """
    `FulfillmentContext` objects are produced by Connectors and are input
    arguments to :meth:`Intent.fulfill`.
    """
    confidence: float
    fulfillment_text: str
    fulfillment_messages: "intents.language.intent_language.IntentResponseDict"
    language: LanguageCode

@dataclass
class FulfillmentResult:
    """
    `FulfillmentResult` are produced by `Intent.fulfill`, and then converted by
    Connectors into Service-actionable responses.

    .. note::

        At the moment this class is only used internally. It is safe to return
        simple :class:`Intent` instances in :meth:`Intent.fulfill`
    """
    trigger: "Intent" = None
    fulfillment_text: List[str] = None
    fulfillment_messages: List["intents.language.intent_language.IntentResponse"] = None
    
    @staticmethod
    def ensure(fulfill_return_value: Union["FulfillmentResult", "Intent"]) -> "FulfillmentResult":
        """
        Convert the given object to a :class:`FulfillmentResult` instance, if it
        is not already one.

        Args:
            fulfill_return_value: An object, as it is returned by :meth:`Intent.fulfill`

        Raises:
            ValueError: If input object is not a valid return value for :meth:`Intent.fulfill`

        Returns:
            FulfillmentResult: A `FulfillmentResult` object representing input
        """
        if fulfill_return_value is None:
            return

        if isinstance(fulfill_return_value, FulfillmentResult):
            return fulfill_return_value

        if isinstance(fulfill_return_value, Intent):
            return FulfillmentResult(trigger=fulfill_return_value)

        raise ValueError(f"Unsupported fulfillment return value: {fulfill_return_value}")

class IntentType(type):

    name: str = None
    lifespan: int = 5

    def __new__(cls, name, bases, dct):
        result_cls = super().__new__(cls, name, bases, dct)

        # Do not process Intent base class
        if name == 'Intent':
            assert not bases
            return result_cls

        for base_cls in bases:
            if base_cls is not Intent:
                if issubclass(base_cls, Intent) and not is_dataclass_strict(base_cls):
                    logger.warning("Intent '%s' is inheriting from non-dataclass '%s'. Did you forget the decorator?", result_cls, base_cls)

        if "name" not in result_cls.__dict__:
            result_cls.name = _intent_name_from_class(result_cls)

        names.check_name(result_cls.name)

        # TODO: check that custom parameters don't overlap Intent fields

        return result_cls

    @property
    def parameter_schema(cls) -> Dict[str, IntentParameterMetadata]:
        if cls is Intent:
            return {}

        if not is_dataclass_strict(cls):
            logger.warning("%s is not a dataclass. This may cause unexpected behavior: consider "
                           "adding a @dataclass decorator to your Intent class.", cls)
            cls = dataclass(cls)

        result = {}
        for param_field in cls.__dataclass_fields__.values():
        # for param_field in cls.__dict__['__dataclass_fields__'].values():
            # List[...]
            if inspect.isclass(param_field.type) and issubclass(param_field.type, Intent):
                continue

            if isinstance(param_field.type, _GenericAlias):
                if param_field.type.__dict__.get('_name') != 'List':
                    raise ValueError(f"Invalid typing '{param_field.type}' for parameter '{param_field.name}'. Only 'List' is supported.")

                if len(param_field.type.__dict__.get('__args__')) != 1:
                    raise ValueError(f"Invalid List modifier '{param_field.type}' for parameter '{param_field.name}'. Must define exactly one inner type (e.g. 'List[Sys.Integer]')")
                
                # From here on, check the inner type (e.g. List[Sys.Integer] -> Sys.Integer)
                entity_cls = param_field.type.__dict__.get('__args__')[0]
                is_list = True
            else:
                entity_cls = param_field.type
                is_list = False

            if not issubclass(entity_cls, entity.EntityMixin):
                raise ValueError(f"Parameter '{param_field.name}' of intent '{cls.name}' is of type '{entity_cls}', which is not an Entity.")

            required = True
            default = None
            if not isinstance(param_field.default, dataclasses._MISSING_TYPE):
                required = False
                default = param_field.default
            if not isinstance(param_field.default_factory, dataclasses._MISSING_TYPE):
                required = False
                default = param_field.default_factory()

            if not required and is_list and not isinstance(default, list):
                raise ValueError(f"List parameter has non-list default value in intent {cls}: {param_field}")

            result[param_field.name] = IntentParameterMetadata(
                name=param_field.name,
                entity_cls=entity_cls,
                is_list=is_list,
                required=required,
                default=default
            )

        return result

class Intent(metaclass=IntentType):
    """
    This is a base class for user defined intents, which are the fundamental
    building blocks of your Agent.

    In its simplest form, an Intent can be defined as follows:

    .. code-block:: python

        from intents import Intent

        class UserSaysHello(Intent):
            \"\"\"A little docstring for my Intent\"\"\"

    *Intents* will then look for language resources in the folder where your
    Agent class is defined, and specifically in
    `language/<LANGUAGE-CODE>/smalltalk.UserSaysHello.yaml`. Note that the
    name of the module where the Intent is defined (`smalltalk.py`) is used as a
    prefix. More details in :mod:`intents.language`.

    Intents can be more complex than this, for instance:

    .. code-block:: python

        from dataclasses import dataclass
        from intents import Intent, Sys

        @dataclass
        class UserSaysHello(Intent):
            \"\"\"A little docstring for my Intent\"\"\"

            user_name: Sys.Person

            name = "hello_custom_name"
            lifespan = 10

    This Intent has a **custom name**, so it will appear as "hello_custom_name" when
    exported to prediction services such as Dialogflow, and its language file
    will just be `hello_custom_name.yaml`, without module prefix. See
    :func:`~intents.model.names.check_name` for more details on the rules that
    intent names must follow.

    This intent has a **lifespan** member. This is used in
    :mod:`intents.model.relations` to determine how long should the Intent stay
    in the conversation context after it is predicted.

    Most importantly, this intent has a `user_name` **parameter** of type
    :class:`Sys.Person` (check out :class:`intents.model.entity.Sys` for
    available system entities). With adequate examples in its language file, it
    will be able to match utterances like "Hello, my name is John", tagging
    "John" as an Entity. When a connector is instantiated, predictions will look
    like this:

    >>> predicted = connector.predict("My name is John")
    >>> predicted.intent
    UserSaysHello(user_name="John")
    >>> predicted.intent.user_name
    "John"
    >>> predicted.fulfillment_text
    "Hi John, I'm Agent"

    .. warning::

        Parameter names are meant to be **unique** within an Agent. That is, the same parameter
        must always be defined with the same type. This is because 1) Some services (e.g. 
        `Alexa <https://developer.amazon.com/en-US/docs/alexa/custom-skills/create-the-interaction-model-for-your-skill.html#intent-slot-names>`_)
        require so. 2) Predictions overwrite parameters in context based on their name: better
        make sure they are compatible.

    Last, we notice the **@dataclass** decorator. We require it to be added
    explicitly, this way IDEs will recognize the Intent class as a dataclass and
    provide correct hints for names and types.
    """
    # TODO: check parameter names: no underscore, no reserved names, max length

    name: str = None

    @property
    def parameter_schema(self) -> Dict[str, IntentParameterMetadata]:
        """
        Return a dict representing the Intent parameter definition. A key is a
        parameter name, a value is a :class:`IntentParameterMetadata` object.

        TODO: this doesn't show up in docs
        """
        return self.__class__.parameter_schema

    def parameter_dict(self) -> Dict[str, Any]:
        result = {}
        for parameter_name in self.parameter_schema.keys():
            result[parameter_name] = getattr(self, parameter_name)
        return result

    @classmethod
    def parent_intents(cls) -> List[Type["Intent"]]:
        """
        Return a list of the Intent parent classes. This wraps `cls.mro()`
        filtering out non-Intent parents (e.g. :class:`object`), :class:`Intent`
        and `cls` itself.

        >>> AskEspresso.parent_intents()
        [AskCoffee, AskDrink]

        Returns:
            A list of parent Intent classes
        """
        result = []
        for c in cls.mro():
            if issubclass(c, Intent) and c is not Intent and c is not cls:
                result.append(c)
        return result

    def fulfill(self,
        context: FulfillmentContext,
        **kwargs
    ) -> "Intent":
        """
        Override this method to define a custom procedure that will be invoked
        when the Intent is predicted, possibly triggering another intent as a
        result. For instance:

        .. code-block:: python

            @dataclass
            class UserConfirmsPayment(Intent):
                
                parent_order: UserOrdersSomething = follow()

                def fulfill(self, context):
                    if shop_api.is_item_available(parent_order.item_id):
                        shop_api.make_order(parent_order.item_id)
                        return OrderSuccessResponse()
                    else:
                        return OrderFailedResponse()
                        
        In the example, when `UserConfirmsPayment` is predicted, it will run
        some business logic, and choose a different response depending on the
        result. Note that response objects are full-fledged Intents, and even
        though they probably won't include example utterances, they can
        define parameters, relations, and their own :meth:`fulfillment`
        implementation, that will be executed recursively.

        More details about fulfillments can be found in the
        :mod:`~intents.fulfillment` module documentation.

        Args:
            context: Context information about the fulfillment request
                (confidence, language, ...)
        Returns:
            A `FulfillmentResult` object, that can contain a followup intent to
            trigger, or just some responses to post back to user.
        """
        return None

def _intent_name_from_class(intent_cls: Type[Intent]) -> str:
    full_name = f"{intent_cls.__module__}.{intent_cls.__name__}"
    if intent_cls.__module__.startswith("_"):
        full_name = intent_cls.__name__
    # full_name = re.sub(r"_+", "_", full_name)
    return ".".join(full_name.split(".")[-2:])



from dataclasses import dataclass

import pytest

from intents import Intent, Sys
from intents.language.intent_language import IntentResponseDict, IntentResponseGroup, TextIntentResponse, QuickRepliesIntentResponse, ImageIntentResponse, CardIntentResponse, CustomPayloadIntentResponse

@dataclass
class FakeIntent(Intent):
    foo: Sys.Person

def test_text_intent_response_string():
    text_response_instance = TextIntentResponse(["ciao"])
    text_response_from_yaml = TextIntentResponse.from_yaml("ciao")
    assert text_response_instance == text_response_from_yaml
    assert text_response_from_yaml.choices == ["ciao"]

def test_text_intent_response_list():
    text_response_instance = TextIntentResponse(["pippo", "franco"])
    text_response_from_yaml = TextIntentResponse.from_yaml(["pippo", "franco"])
    assert text_response_instance == text_response_from_yaml
    assert text_response_from_yaml.choices == ["pippo", "franco"]

def test_text_intent_response_equals():
    response_1 = TextIntentResponse(["pippo", "franco"])
    response_2 = TextIntentResponse(["pippo", "franco"])
    response_3 = TextIntentResponse(["franco", "pippo"])
    response_4 = TextIntentResponse(["something", "else"])
    assert response_1 == response_2
    assert response_1 != response_3
    assert response_1 != response_4
    # can't do this with mutable lists
    # assert response_1 in {response_2, response_3, response_4}
    # assert response_1 not in {response_3, response_4}

def test_text_intent_response_random():
    text_response_instance = TextIntentResponse(["pippo", "franco"])
    assert text_response_instance.random() in ["pippo", "franco"]

def test_text_response_render():
    text_response_instance = TextIntentResponse([
        "$foo is when you look at me",
        "I like it when it's $foo",
        "Why does it $foo when I $foobar?",
        "Why does it $foobar when I $foo?",
        "Is it \"$foo\" if it hurts?"
        "$foo"
    ])

    fake_intent = FakeIntent(foo="Trotskyism")

    expected = TextIntentResponse([
        "Trotskyism is when you look at me",
        "I like it when it's Trotskyism",
        "Why does it Trotskyism when I $foobar?",
        "Why does it $foobar when I Trotskyism?",
        "Is it \"Trotskyism\" if it hurts?"
        "Trotskyism"
    ])

    assert text_response_instance.render(fake_intent) == expected

def test_quick_replies_response_string():
    quick_replies_instance = QuickRepliesIntentResponse(["ciao"])
    quick_replies_from_yaml = QuickRepliesIntentResponse.from_yaml("ciao")
    assert quick_replies_instance == quick_replies_from_yaml
    assert quick_replies_from_yaml.replies == ["ciao"]

def test_quick_replies_render():
    quick_replies_response_instance = QuickRepliesIntentResponse(["$foo", "pippo", "franco $foo"])
    fake_intent = FakeIntent(foo="bar")
    expected = QuickRepliesIntentResponse(["bar", "pippo", "franco bar"])

def test_quick_replies_response_list():
    quick_replies_response_instance = QuickRepliesIntentResponse(["pippo", "franco"])
    quick_replies_response_from_yaml = QuickRepliesIntentResponse.from_yaml(["pippo", "franco"])
    assert quick_replies_response_instance == quick_replies_response_from_yaml
    assert quick_replies_response_from_yaml.replies == ["pippo", "franco"]

def test_quick_replies_response_too_long():
    with pytest.raises(ValueError):
        QuickRepliesIntentResponse(["pippo", "franco", "chi chi chi, co co co"])
        
    with pytest.raises(ValueError):
        QuickRepliesIntentResponse.from_yaml(["chi chi chi, co co co", "pippo", "franco"])

    with pytest.raises(ValueError):
        QuickRepliesIntentResponse.from_yaml("chi chi chi, co co co")

def test_image_response_string():
    image_response_instance = ImageIntentResponse(url="https://example.com/image.png")
    image_response_from_yaml = ImageIntentResponse.from_yaml("https://example.com/image.png")
    assert image_response_instance == image_response_from_yaml

def test_image_response_dict():
    image_response_instance = ImageIntentResponse(url="https://example.com/image.png", title="A title")
    image_response_from_yaml = ImageIntentResponse.from_yaml({"url": "https://example.com/image.png", "title": "A title"})
    assert image_response_instance == image_response_from_yaml

def test_custom_payload_response():
    custom_payload_instance = CustomPayloadIntentResponse("any_name", {"foo": "bar"})
    custom_payload_from_yaml = CustomPayloadIntentResponse.from_yaml({
        "any_name": {"foo": "bar"}
    })
    assert custom_payload_instance == custom_payload_from_yaml

def test_card_intent_response():
    card_reponse_instance = CardIntentResponse("Card Title", "a subtitle...")
    card_reponse_from_yaml = CardIntentResponse.from_yaml({
        "title": "Card Title",
        "subtitle":"a subtitle..."
    })
    assert card_reponse_instance == card_reponse_from_yaml

def test_custom_payload_not_dict_data():
    with pytest.raises(ValueError):
        CustomPayloadIntentResponse.from_yaml("foo")

def test_custom_payload_not_dict_value():
    with pytest.raises(ValueError):
        CustomPayloadIntentResponse.from_yaml({"foo": "bar"})

def test_custom_payload_multiple_keys():
    with pytest.raises(ValueError):
        CustomPayloadIntentResponse.from_yaml({
            "any_name": {"foo": "bar"},
            "another_name": {"not": "allowed"}
        })

def test_intent_response_dict_for_group():
    mock_default_messages = ["MOCK DEFAULT 1", "MOCK DEFAULT 2"]
    mock_rich_messages = ["MOCK RICH 1", "MOCK RICH 2"]
    d = IntentResponseDict({
        IntentResponseGroup.DEFAULT: mock_default_messages,
        IntentResponseGroup.RICH: mock_rich_messages
    })
    assert d.for_group() == mock_rich_messages
    assert d.for_group(IntentResponseGroup.DEFAULT) == mock_default_messages
    assert d.for_group(IntentResponseGroup.RICH) == mock_rich_messages



"""
This module defines general purpose helpers that are used throughout the project
"""
import dataclasses
from enum import Enum
from datetime import datetime
from dataclasses import asdict, field

class CustomFields(Enum):
    OMIT_NONE = "OMIT_NONE"

def OmitNone():
    return field(default=CustomFields.OMIT_NONE)

def custom_asdict_factory():
    """
    Return a custom dict factory to use with dataclasses' :func:`asdict`. Custom
    behavior include:

    * Serialize Enums with their value
    * Omit fields whose value is OmitNone() (e.g. `foo: str = OmitNone()`)

    This is mainly used to serialize schemas in Connectors.

    source: https://stackoverflow.com/questions/61338539/how-to-use-enum-value-in-asdict-function-from-dataclasses-module

    TODO: add optional "exclude_none" to exclude None fields from output
    """

    def result_f(data):

        def convert_value(obj):
            if isinstance(obj, Enum):
                return obj.value
            if isinstance(obj, datetime):
                return str(obj)
            return obj

        return dict((k, convert_value(v)) for k, v in data if v != CustomFields.OMIT_NONE)
    
    return result_f

def is_dataclass_strict(obj):
    """
    Like :func:`dataclasses.is_dataclass`, but return True only if the class
    itself was decorated with `@dataclass` (that is, inheriting from a parent
    dataclass is not sufficient)

    .. code-block:: python

        @dataclass
        class a_class:
            foo: str

        class a_subclass(a_class):
            bar: str

        is_dataclass(a_subclass)           # True
        is_dataclass_strict(a_subclass)    # False
    """
    cls = obj if isinstance(obj, type) else type(obj)
    return dataclasses._FIELDS in cls.__dict__

def to_dict(dataclass_obj):
    return asdict(dataclass_obj, dict_factory=custom_asdict_factory())



import pytest

from intents import Sys, Agent, Entity
from intents.language import LanguageCode
from intents.connectors.interface import Connector, StringEntityMapping, ServiceEntityMappings

class DummyConnector(Connector):

    entity_mappings = ServiceEntityMappings.from_list([
        StringEntityMapping(Sys.Person, "fake-person-service-name")
    ])

    def predict(self, message, session, language):
        return None
    
    def trigger(self, intent, session, language):
        return None
    
    def fulfill(self, fulfillment_request):
        return None
        
    def upload(self):
        return None

    def export(self, destination):
        return None

class DummyAgent(Agent):
    languages = ['it', 'en']

def test_default_session_is_not_none():
    connector = DummyConnector(DummyAgent)
    assert connector.default_session

def test_language_code_string_is_cast():
    connector = DummyConnector(DummyAgent, default_language="en")
    assert connector.default_language == LanguageCode.ENGLISH

    connector = DummyConnector(DummyAgent, default_language=LanguageCode.ENGLISH)
    assert connector.default_language == LanguageCode.ENGLISH

def test_first_language_is_default():
    connector = DummyConnector(DummyAgent)
    assert connector.default_language == LanguageCode.ITALIAN

def test_entity_mappings__system_entity():
    m = StringEntityMapping(Sys.Person, "fake-person-service-name")
    mappings = ServiceEntityMappings.from_list([m])

    assert mappings.lookup(Sys.Person) is m
    assert mappings.service_name(Sys.Person) == "fake-person-service-name"

def test_entity_mappings__custom_entity():
    mappings = ServiceEntityMappings()

    class MockCustomEntity(Entity):
        pass

    class MockCustomEntityWithName(Entity):
        name = "my_entity_name"

    assert mappings.lookup(MockCustomEntity) == StringEntityMapping(MockCustomEntity, "MockCustomEntity")
    assert mappings.lookup(MockCustomEntityWithName) == StringEntityMapping(MockCustomEntityWithName, "my_entity_name")

    assert mappings.service_name(MockCustomEntity) == "MockCustomEntity"
    assert mappings.service_name(MockCustomEntityWithName) == "my_entity_name"

def test_entity_mappings__key_error():
    m = StringEntityMapping(Sys.Person, "fake-person-service-name")
    mappings = ServiceEntityMappings.from_list([m])

    class MockCustomEntity(Entity):
        pass

    with pytest.raises(KeyError):
        mappings.lookup(Sys.PhoneNumber)

    with pytest.raises(KeyError):
        mappings.service_name(Sys.PhoneNumber)



from dataclasses import dataclass
from intents import Intent
from intents.model.relations import follow, intent_relations, IntentRelationMap, IntentRelation, RelationType, FollowRelationParameters, FollowIntentRelation

def test_intent_relations__no_relations():
    
    @dataclass
    class NoRelationsIntent(Intent):
        pass

    result = intent_relations(NoRelationsIntent)
    assert isinstance(result, IntentRelationMap)
    assert result.follow == []

def test_intent_relations__follow():
    
    @dataclass
    class FollowedIntent(Intent):
        pass

    @dataclass
    class FollowingIntent(Intent):
        parent_followed_intent: FollowedIntent = follow()

    result = intent_relations(FollowingIntent)
    expected = [
        FollowIntentRelation(
            relation_parameters=FollowRelationParameters(new_lifespan=None),
            field_name='parent_followed_intent',
            source_cls=FollowingIntent,
            target_cls=FollowedIntent
        )
    ]
    assert result.follow == expected
    assert result.follow[0].relation_type == RelationType.FOLLOW

def test_subclass_inherits_follow():

    @dataclass
    class FollowedIntent(Intent):
        pass

    @dataclass
    class FollowingIntent(Intent):
        parent_followed_intent: FollowedIntent = follow()

    @dataclass
    class FollowingIntentSubclass(FollowingIntent):
        pass

    result = intent_relations(FollowingIntentSubclass)
    expected = [
        FollowIntentRelation(
            relation_parameters=FollowRelationParameters(new_lifespan=None),
            field_name='parent_followed_intent',
            source_cls=FollowingIntentSubclass,
            target_cls=FollowedIntent
        )
    ]
    assert result.follow == expected
    assert result.follow[0].relation_type == RelationType.FOLLOW



"""
Patches :class:`~intents.model.entity.Sys.Color`

Sources:

* https://simple.wikipedia.org/wiki/List_of_colors
* https://it.wikipedia.org/wiki/Lista_dei_colori
* https://fr.wikipedia.org/wiki/Liste_de_noms_de_couleur
* https://de.wikipedia.org/wiki/Liste_der_Farben_im_Farbkreis
* https://www.lingomastery.com/full-list-colors-spanish/

Some uncommons colors have been removed
"""
from typing import List, Union

from intents import Entity, LanguageCode
from intents.language.entity_language import make_language_data

# TODO: Dutch, German and Spanish have less than other languages

class I_IntentsColor(Entity):
    __intents_internal__ = True
    __entity_language_data__ = {
        LanguageCode.ENGLISH: make_language_data(["Amaranth", "Amber", "Amethyst", "Apricot", "Aquamarine", "Azure", "Baby blue", "Beige", "Brick red", "Black", "Blue", "Blue-green", "Blue-violet", "Blush", "Bronze", "Brown", "Burgundy", "Byzantium", "Carmine", "Cerise", "Cerulean", "Champagne", "Chartreuse green", "Chocolate", "Cobalt blue", "Coffee", "Copper", "Coral", "Crimson", "Cyan", "Desert sand", "Electric blue", "Emerald", "Erin", "Gold", "Gray", "Green", "Harlequin", "Indigo", "Ivory", "Jade", "Jungle green", "Lavender", "Lemon", "Lilac", "Lime", "Magenta", "Magenta rose", "Maroon", "Mauve", "Navy blue", "Ochre", "Olive", "Orange", "Orange-red", "Orchid", "Peach", "Pear", "Periwinkle", "Persian blue", "Pink", "Plum", "Prussian blue", "Puce", "Purple", "Raspberry", "Red", "Red-violet", "Rose", "Ruby", "Salmon", "Sangria", "Sapphire", "Scarlet", "Silver", "Slate gray", "Spring bud", "Spring green", "Tan", "Taupe", "Teal", "Turquoise", "Ultramarine", "Violet", "Viridian", "White", "Yellow"]),
        LanguageCode.ITALIAN: make_language_data(["Acquamarina", "Albicocca", "Amaranto", "Ambra", "Ametista", "Anguria", "Antracite", "Aragosta", "Arancione", "Ardesia", "Argento", "Asparago", "Avio", "Avorio", "Azalea", "Azzurro", "Beige", "Beige-oliva", "Beige verdastro", "Bianco", "Bianco fantasma", "Bianco fumo", "Bianco Navajo", "Biscotto", "Bistro", "Blu", "Blu acciaio", "Blu alice", "Blu Bondi", "Blu cadetto", "Blu ceruleo", "Blu di Persia", "Blu di Prussia", "Blu elettrico", "Blu marino", "Blu notte", "Blu oltremare", "Blu polvere", "Blu reale", "Blu scuro", "Bordeaux", "Borgogna", "Bronzo", "Bronzo antico", "Camoscio", "Carbone", "Cardo", "Carminio", "Carta da zucchero", "Castagno", "Catrame", "Celadon", "Celeste", "Ceruleo", "Chartreuse", "Ciano", "Ciliegia", "Cioccolato", "Cobalto", "Conchiglia", "Corallo", "Crema", "Cremisi", "Denim", "Eliotropo", "Écru", "Fiore di granturco", "Foglia di tè", "Fucsia", "Fulvo", "Gainsboro", "Giada", "Giallo", "Giallo miele", "Giallo pastello", "Giallo sabbia", "Glicine", "Granata", "Grano", "Grigio ardesia", "Grigio cenere", "Grigio topo", "Indaco", "Indaco elettrico", "Indaco scuro", "Isabella", "Kaki", "Lampone", "Lavanda", "Lavanda pallido", "Lavanda rosata", "Limone", "Limone crema", "Lilla", "Lime", "Lino", "Magenta", "Magnolia", "Malva", "Malva chiaro", "Mandarino", "Marrone", "Marrone chiaro", "Marrone pastello", "Marrone-rosso", "Marrone scuro", "Melanzana", "Mogano", "Nero", "Ocra", "Olivina", "Orchidea", "Oro", "Oro vecchio", "Ottone antico", "Ottanio", "Papaia", "Pera", "Pervinca", "Pesca", "Pesca scuro", "Pesca-arancio", "Pesca-giallo", "Pistacchio", "Platino", "Porpora", "Prugna", "Rame", "Rosa", "Rosa arancio", "Rosa medio", "Rosa pallido", "Rosa pastello", "Rosa scuro", "Rosa shocking", "Rosa vivo", "Rosso", "Rosso aragosta", "Rosso cardinale", "Rosso corsa", "Rosso Falun", "Rosso fragola", "Rosso fuoco", "Rosso mattone", "Rosso pomodoro", "Rosso pompeiano", "Rosso rosa", "Rosso sangue", "Rosso segnale", "Rosso Tiziano", "Rosso veneziano", "Rubino", "Sabbia", "Salmone", "Scarlatto", "Senape", "Seppia", "Solidago", "Tanno", "Tenné", "Tronco", "Turchese", "Turchese chiaro", "Turchese pallido", "Turchese scuro", "Verde", "Verde foresta", "Verde chiaro", "Verde-giallo", "Verde marino", "Verde menta", "Verde muschio", "Verde oliva", "Verde olivastro", "Verde pastello", "Verde pino", "Verde primavera", "Verde ufficio", "Verde smeraldo", "Verde Veronese", "Vermiglio", "Viola", "Viola chiaro", "Viola melanzana", "Violetto", "Vinaccia", "Zafferano", "Zaffiro"]),
        LanguageCode.DUTCH: make_language_data(["zwart", "wit", "rood", "geel", "blauw", "groen", "bruin", ["rose", "roze"], "oranje", "grijs", ["paars", "purper"], "lila", "turquoise", "khaki", "gebroken wit", ["okergeel", "oker"], "aubergine", "bordeauxrood"]),
        LanguageCode.FRENCH: make_language_data(["Abricot", "Acajou", "Aigue-marine", "Amande", "Amarante", "Ambre", "Améthyste", "Anthracite", "Argent", "Aubergine", "Auburn", "Aurore", "Avocat", "Azur", "Baillet", "Beurre", "Bis", "Bisque", "Bistre", "Bitume", "Blanc", "Blé", "Bleu", "Bleuet", "Bleu acier", "Bleu barbeau", "Bleu céleste", "Bleu ciel", "Bleu de cobalt", "Bleu électrique", "Bleu givré", "Bleu marine", "Bleu nuit", "Bleu pétrole", "Bleu turquin", "Blond vénitien", "Blond", "Brique", "Bronze", "Brou de noix", "Caca d'oie", "Cacao", "Cachou", "Cæruleum", "Café", "Cannelle", "Capucine", "Caramel", "Carmin", "Carotte", "Chamois", "Chartreuse", "Châtain", "Chaudron", "Chocolat", "Cinabre", "Citrouille", "Corail", "Cramoisi", "Cuisse de nymphe", "Cuivre", "Cyan", "Écarlate", "Écru", "Émeraude", "Fauve", "Flave", "Fraise", "Framboise", "Fuchsia", "Fumée", "Garance", "Glauque", "Glycine", "Grège", "Grenadine", "Grenat", "Gris", "Gris fer", "Gris perle", "Gris souris", "Groseille", "Gueules", "Incarnat", "Indigo", "Isabelle", "Jaune", "Jaune canari", "Jaune citron", "Jaune d'or", "Jaune mimosa", "Lapis-lazuli", "Lavande", "Lilas", "Lime", "Lin", "Magenta", "Mandarine", "Marron", "Mastic", "Mauve", "Menthe", "Moutarde", "Nacarat", "Nankin", "Noisette", "Noir", "Ocre jaune", "Ocre rouge", "Olive", "Or", "Orange", "Orchidée", "Orpiment", "Paille", "Parme", "Pelure d'oignon", "Pervenche", "Pistache", "Poil de chameau", "Ponceau ou Coquelicot", "Pourpre", "Prasin", "Prune", "Puce", "Rose", "Rouge", "Rouge d'Andrinople", "Rouge de Falun", "Rouge feu", "Rouille", "Roux", "Rubis", "Sable", "Safre", "Sanguine", "Saphir", "Sarcelle", "Saumon", "Sépia", "Smalt", "Soufre", "Tabac", "Taupe", "Terre d'ombre", "Tomate", "Topaze", "Tourterelle", "Turquoise", "Vanille", "Vermeil", "Vermillon", "Vert", "Vert bouteille", "Vert céladon", "Vert d'eau", "Vert de Hooker", "Vert olive", "Violet", "Viride", "Zinzolin"]),
        LanguageCode.GERMAN: make_language_data(["Rot", "Gelb", "Grün", "Blau", "Braun", "Mittelrot", "Zinnoberrot", "Rotorange", "Ocker", "Gelborange", "Indischgelb", "Umbra", "Goldgelb", "Cadmiumgelb", "Mittelgelb", "Zitronengelb", "Sepia", "Maigrün", "Gelbgrün", "Grasgrün", "Olivgrün", "Blattgrün", "Mittelgrün", "Blaugrün", "Smaragdgrün", "Phthalogrün", "Seegrün", "Türkis", "Türkisblau", "Cyanblau", "Coelinblau", "Bremerblau", "Indigo", "Preußischblau", "Mittelblau", "Pariserblau", "Ultramarinblau", "Paynesgrau", "Kobaltblau", "Violettblau", "Blauviolett", "Violett", "Purpurviolett", "Purpurrot", "Krapplack", "Karminrot", "Venezianischrot", "Scharlachrot"]),
        LanguageCode.SPANISH: make_language_data(["Negro", "Blanco", "Amarillo", "Azul", "Rojo", "Verde", "Naranja", "Violeta/Morado", "Púrpura", "Rosa", "Gris", "Marrón", "Plateado/Plata", "Dorado/Oro", "Limón", "Lima", "Mostaza", "Lavanda", "Lila", "Salmón", "Fucsia", "Magenta", "Escarlata", "Borgoña", "Marfil", "Champaña", "Beige", "Ocre", "Cobre", "Ámbar", "Aguamarina", "Turquesa", "Cian", "Azul Claro", "Azul de Prusia", "Azul Marino", "Jade", "Verde Oliva"]),
        LanguageCode.CHINESE: make_language_data(["黑色", "灰色", "銀色", "白煙色", "白色", "雪色", "鐵灰色", "沙棕", "印度紅", "褐色", "磚紅色", "栗色", "紅色", "柿子橙", "鮭紅", "腥紅", "蕃茄紅", "珊瑚紅", "橙紅", "朱紅", "赭黃", "駝色", "杏黃", "海貝色", "鞍褐", "巧克力色", "粉撲桃色", "沙褐", "銅色", "亞麻色", "烏賊墨色", "赭色", "陶坯黃", "橘色", "古董白", "日曬色", "硬木色", "杏仁白", "那瓦霍白", "萬壽菊黃", "灰土色", "卡其色", "鹿皮鞋色", "舊蕾絲色", "小麥色", "桃色", "橙色", "金菊色", "咖啡色", "茉莉黃", "琥珀色", "鉻黃", "金色", "灰金菊色", "含羞草黃", "奶油色", "象牙色", "米色", "香檳黃", "芥末黃", "月黃", "橄欖色", "鮮黃", "黃色", "苔蘚綠", "亮檸檬绿", "橄欖軍服綠", "黃綠", "暗橄欖綠", "蘋果綠", "綠黃", "草綠", "草坪綠", "葉綠", "嫩綠", "明綠", "鈷綠", "蜜瓜綠", "暗海綠", "亮綠", "灰綠", "常春藤綠", "森林綠", "檸檬綠", "暗綠", "綠色", "鮮绿色", "薄荷綠", "青瓷綠", "綠松石綠", "鉻綠", "海綠", "中海綠", "薄荷奶油色", "孔雀綠", "碧藍色", "青藍", "水藍", "土耳其藍", "綠松石色", "中綠松石色", "亮青", "淺藍", "鳧綠", "青色", "水色", "孔雀藍", "薩克斯藍", "深天藍", "水手藍", "道奇藍", "礦藍", "湛藍", "韋奇伍德瓷藍", "鈷藍", "灰丁寧藍", "鼠尾草藍", "藍寶石色", "蔚藍", "暗礦藍", "極濃海藍", "天青石藍", "薰衣草紫", "長春花色", "長春花色", "午夜藍", "藏青", "藍色", "淺灰紫紅", "靛色", "鋼青色", "碧綠", "紫色", "岩藍", "紫羅蘭色", "岩灰", "藍紫", "梅紅色", "天空藍", "淺天藍", "勃艮第酒紅", "宗主教", "紫水晶色", "薰衣草紫", "三色堇紫", "錦葵紫", "優品紫紅", "淡紫丁香色", "礦紫", "紫藤色", "紫丁香色", "纈草紫", "木槿紫", "薊紫", "鐵線蓮紫", "洋紅", "品紅", "蘭紫", "淺珍珠紅", "陳玫紅", "淺玫瑰紅", "洋玫瑰紅", "玫瑰紅", "紅寶石色", "山茶紅", "火鶴紅", "淺珊瑚紅", "尖晶石紅", "胭脂紅", "淺粉紅", "樞機紅", "薰衣草紫紅", "櫻桃紅", "淺鮭紅", "緋紅", "粉紅", "殼黃紅"]),
    }



import logging

from intents.connectors._experimental.alexa.connector import AlexaConnector

logger = logging.getLogger(__name__)
logger.warning("Alexa support is EXPERIMENTAL. Features may be limited, inconsistent, or change without notice.")



"""
The :class:`Agent` base class is the entry point of your project. You will
subclass it when defining your own Agent, and will later :meth:`Agent.register`
Intent classes and other resources into it.

Once the Agent is defined, you will connect it to a cloud service with a
:class:`Connector`, to make prediction and trigger requests.
"""
import re
import logging
from types import ModuleType
from dataclasses import dataclass
from inspect import isclass, getmembers
from typing import List, Dict, Union, Set, Type

from intents import LanguageCode
from intents.model.intent import Intent, IntentParameterMetadata
from intents.model.entity import EntityMixin, SystemEntityMixin
from intents import language

logger = logging.getLogger(__name__)

@dataclass
class RegisteredParameter:
    """
    Agents register intent parameters. If two intents declare the same parameter
    name, they must also declare the same type for it. 
    """
    metadata: IntentParameterMetadata
    used_in: List[type(Intent)]

class AgentType(type):

    languages: List[LanguageCode] = None

    intents: List[Intent] = None
    _intents_by_name: Dict[str, Intent] = None
    _intents_by_norm_name: Dict[str, Intent] = None # my_module.HelloWorld -> mymodulehelloworld
    _entities_by_name: Dict[str, Type[EntityMixin]] = None
    _parameters_by_name: Dict[str, RegisteredParameter] = None
    _referenced_sys_entities: Set[SystemEntityMixin] = None # All

    def __new__(cls, name, bases, dct):
        result_cls = super().__new__(cls, name, bases, dct)
        
        if name == 'Agent':
            assert not bases
            return result_cls

        if not result_cls.languages:
            result_cls.languages = language.agent_supported_languages(result_cls)

        languages = []
        for lan in result_cls.languages:
            if isinstance(lan, LanguageCode):
                languages.append(lan)
            elif isinstance(lan, str):
                languages.append(LanguageCode(lan))
            else:
                raise ValueError(f"Unsupported language '{lan}' for Agent '{result_cls}'. Must be a value of 'intents.LanguageCode'")
        result_cls.languages = languages

        return result_cls

class Agent(metaclass=AgentType):
    """
    As the name suggests, Agent is the base class that models an Agent
    definition within the *Intents* framework.

    Typically, you will define a single Agent class in your project, that could
    be as simple as this:

    .. code-block:: python

        from intents import Agent

        class MyAgent(Agent):
            \"\"\"A little docstring for your Agent\"\"\"

    You can optionally define the languages that you intend to support.
    *Intents* will look for language resources based on the `language` class
    variable:

    .. code-block:: python

        class MyAgent(Agent):
            \"\"\"A little docstring for your Agent\"\"\"
            languages = ["en", "it"]

    Languages are values from :class:`LanguageCode`. If omitted,
    *Intents* will discover language resources by itself.

    You won't do much more with your *Agent* class, other than registering
    intents and resources with :meth:`Agent.register`, or passing it to a
    :class:`Connector` to make predictions.
    """

    @classmethod
    def register(cls, resource: Union[Type[Intent], ModuleType]):
        """
        Register the given resource in Agent. The resource could be:

        * An :class:`~intents.model.intent.Intent`
        * A module. In this case, the module is scanned (non recursively) for
          Intents, and each Intent is added individually

        This is how you register a **single intent**:

        .. code-block:: python

            from intents import Agent, Intent

            class MyAgent(Agent):
                pass

            @dataclass
            class MyTestIntent(Intent):
                \"\"\"A little docstring for my Intent...\"\"\"
                a_parameter: Sys.Date
                another_parameter: Sys.Person

            MyAgent.register(MyTestIntent)

        Alternatively, you can register a **whole module** containing Intents.
        This is how you register all the intents that are defined in the
        `smalltalk` module of `example_agent`:

        .. code-block:: python

            from example_agent import smalltalk

            class MyAgent(Agent):
                pass

            MyAgent.register(smalltalk)

        Note that together with each Intent, the Agent will register all of its
        linked properties and resources, such as parameters and Entities.

        :param resource: The resource to register (an Intent, or a module
                         containing Intents)
        """
        if isclass(resource) and issubclass(resource, Intent):
            cls._register_intent(resource)

        elif isinstance(resource, ModuleType):
            for member_name, member in getmembers(resource, isclass):
                if member.__module__ == resource.__name__ and issubclass(member, Intent):
                    cls._register_intent(member)

    @classmethod
    def _register_intent(cls, intent_cls: Type[Intent]):
        """
        Register a single intent in the Agent class and check that language data is
        present for all supported languages (examples and responses).
        """
        if not cls.intents or not cls._intents_by_name:
            assert not cls.intents and not cls._intents_by_name
            cls.intents = []
            cls._intents_by_name = {}
            cls._intents_by_norm_name = {}
            cls._entities_by_name = {}
            cls._parameters_by_name = {}
            cls._referenced_sys_entities = set()

        norm_name = Agent._norm_name(intent_cls.name)
        if cls._intents_by_norm_name.get(norm_name):
            raise ValueError(f"Another intent exists with an equivalent name to {intent_cls.name}" +
                f": {cls._intents_by_norm_name[norm_name]}")

        language.intent_language_data(cls, intent_cls) # TODO: Agent languages only

        for param_name, param_metadata in intent_cls.parameter_schema.items():
            cls._register_parameter(param_metadata, intent_cls)
            cls._register_entity(param_metadata.entity_cls, param_name, intent_cls.name)

        cls.intents.append(intent_cls)
        cls._intents_by_name[intent_cls.name] = intent_cls
        cls._intents_by_norm_name[norm_name] = intent_cls

    @classmethod
    def _register_parameter(cls, param_meta: IntentParameterMetadata, intent_cls: type(Intent)):
        existing_param: RegisteredParameter = cls._parameters_by_name.get(param_meta.name)
        if not existing_param:
            cls._parameters_by_name[param_meta.name] = RegisteredParameter(
                metadata=param_meta,
                used_in=[intent_cls]
            )
            return

        if param_meta.entity_cls != existing_param.metadata.entity_cls:
            raise ValueError(f"Parameters with the same name must have the same type. Parameter '{param_meta.name}' " +
                f"is declared in Intent '{intent_cls.name}' with type '{param_meta.entity_cls}'; however, it was " +
                f"also declared in Intents '{existing_param.used_in}' with type '{existing_param.metadata.entity_cls}'")
                
        if param_meta.is_list and not existing_param.metadata.is_list:
            raise ValueError(f"Parameters with the same name must have the same type. Parameter '{param_meta.name}' " +
                f"is declared in Intent '{intent_cls.name}' as a List; however, it was also declared in Intents " +
                f"'{existing_param.used_in}' as not a List")

        existing_param.used_in.append(intent_cls)

    @classmethod
    def _register_entity(cls, entity_cls: Type[EntityMixin], parameter_name: str, intent_name: str):
        if not issubclass(entity_cls, EntityMixin):
            raise ValueError(f"Invalid type '{entity_cls}' for parameter '{parameter_name}' in Intent '{intent_name}': must be an Entity. Try system entities such as 'intents.Sys.Integer', or define your own custom entity.")

        if issubclass(entity_cls, SystemEntityMixin):
            cls._referenced_sys_entities.add(entity_cls)
            return

        existing_cls = cls._entities_by_name.get(entity_cls.name)
        if not existing_cls:
            language.entity_language_data(cls, entity_cls) # Checks that language data is existing and consistent
            cls._entities_by_name[entity_cls.name] = entity_cls
            return

        if id(entity_cls) != id(existing_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            entity_cls_path = f"{entity_cls.__module__}.{entity_cls.__qualname__}"
            raise ValueError(f"Two different Entity classes exist with the same name: '{existing_cls_path}' and '{entity_cls_path}'")

    def _pylint_hack(self):
        raise NotImplementedError()

    @staticmethod
    def _norm_name(intent_name: str) -> str:
        return re.sub(r"[^a-z0-9]", "", intent_name.lower())



"""
Here we define schemas to parse Alexa fulfillment requests into dataclasses
"""
from enum import Enum
from datetime import datetime
from typing import List, Dict, Any, Union
from dataclasses import dataclass, field

import dacite

from intents.helpers.data_classes import OmitNone


class IntentConfirmationStatus(Enum):
    NONE = "NONE"
    CONFIRMED = "CONFIRMED"
    DENIED = "DENIED"

class SlotType(Enum):
    SIMPLE = "Simple"
    LIST = "List"

@dataclass
class IntentSlotValueResolutionItemValueValue:
    id: str
    name: str

@dataclass
class IntentSlotValueResolutionItemValue:
    value: IntentSlotValueResolutionItemValueValue # ☠

@dataclass
class IntentSlotValueResolutionItem:
    authority: str
    status: dict # TODO: model
    values: List[IntentSlotValueResolutionItemValue]

@dataclass
class IntentSlotValueResolution:
    resolutionsPerAuthority: List[IntentSlotValueResolutionItem]

@dataclass
class IntentSlotValue:
    type: SlotType
    value: str = None # Only included when type=SIMPLE
    values: List["IntentSlotValue"] = None # Only included when type=LIST
    resolutions: IntentSlotValueResolution = None

@dataclass
class IntentSlot:
    confirmationStatus: IntentConfirmationStatus
    name: str
    # The following seem to be None when slot is returned but not matched
    source: str = None # always USER...
    slotValue: IntentSlotValue = None
    value: str = None
    resolutions: dict = None # model


#############
# RESPONSE  #
#############

class OutputSpeechType(Enum):
    PlainText = "PlainText"
    SSML = "SSML"

class PlayBehavior(Enum):
    ENQUEUE = "ENQUEUE"
    REPLACE_ALL = "REPLACE_ALL"
    REPLACE_ENQUEUED = "PlainText"

@dataclass
class FulfillmentResponseOutputSpeech:
    type: OutputSpeechType
    text: str = OmitNone()
    ssml: str = OmitNone()
    playBehavior = PlayBehavior

@dataclass
class FulfillmentResponseDirective:
    type: str

@dataclass
class FulfillmentResponseDialogDelegateUpdatedIntent:
    name: str
    confirmationStatus: IntentConfirmationStatus
    slots: Dict[str, IntentSlot]

@dataclass
class FulfillmentResponseDialogDelegateDirective(FulfillmentResponseDirective):
    type: str = "Dialog.Delegate"
    updatedIntent: FulfillmentResponseDialogDelegateUpdatedIntent=None

@dataclass
class FulfillmentResponse:
    outputSpeech: FulfillmentResponseOutputSpeech = None
    card: dict = OmitNone()    # TODO: model
    reprompt: dict = OmitNone()    # TODO: model
    shouldEndSession: bool = True
    directives: List[FulfillmentResponseDirective] = OmitNone() # TODO: model

@dataclass
class FulfillmentResponseBody:
    version: str = "1.0"
    sessionAttributes: dict = field(default_factory=dict)
    response: FulfillmentResponse=OmitNone()

#############
# REQUEST   #
#############

#
# Session
#

@dataclass
class FulfillmentSessionApplication:
    applicationId: str

@dataclass
class FulfillmentSessionUser:
    userId: str
    accessToken: str = None
    permissions: dict = None # deprecated: https://developer.amazon.com/en-US/docs/alexa/custom-skills/request-and-response-json-reference.html#session-object

@dataclass
class FulfillmentSession:
    new: bool
    sessionId: str
    application: FulfillmentSessionApplication
    user: FulfillmentSessionUser
    attributes: Dict[str, Any] = field(default_factory=dict)

#
# Context
# https://developer.amazon.com/en-US/docs/alexa/custom-skills/request-and-response-json-reference.html#context-object
#

class PlayerActivity(Enum):
    IDLE = "IDLE"
    PAUSED = "PAUSED"
    PLAYING = "PLAYING"
    BUFFER_UNDERRUN = "BUFFER_UNDERRUN"
    FINISHED = "FINISHED"
    STOPPED = "STOPPED"

@dataclass
class FulfillmentContextAudioPlayer:
    token: str
    playerActivity: PlayerActivity
    offsetInMilliseconds: float = None

# ------

@dataclass
class FulfillmentContextSystemApplication:
    applicationId: str

@dataclass
class FulfillmentContextSystemDevice:
    deviceId: str
    supportedInterfaces: Dict[str, dict] # TODO: model with Enum

@dataclass
class FulfillmentContextSystemUnit:
    unitId: str
    persistentUnitId: str

@dataclass
class FulfillmentContextSystemPerson:
    personId: str
    accessToken: str = None

@dataclass
class FulfillmentContextSystemUser:
    userId: str
    accessToken: str = None
    permissions: dict = None # deprecated

@dataclass
class FulfillmentContextSystem:
    apiAccessToken: str
    apiEndpoint: str
    application: FulfillmentContextSystemApplication # TODO: use this to verify that request is authentic
    device: FulfillmentContextSystemDevice
    user: FulfillmentContextSystemUser            # the person who owns the account that installed the skill
    person: FulfillmentContextSystemPerson = None # the person that is talking to Alexa (recognized by voice)
    unit: FulfillmentContextSystemUnit = None
    
# -----
# https://developer.amazon.com/en-US/docs/alexa/alexa-presentation-language/apl-interface.html#viewport-properties

class ViewportMode(Enum):
    HUB = "HUB"
    TV = "TV"
    PC = "PC"
    MOBILE = "MOBILE"
    AUTO = "AUTO"

class ViewportShape(Enum):
    ROUND = "ROUND"
    RECTANGLE = "RECTANGLE"

@dataclass
class FulfillmentContextViewportExperience:
    canRotate: bool
    canResize: bool

@dataclass
class FulfillmentContextViewport:
    experiences: List[FulfillmentContextViewportExperience]
    mode: ViewportMode
    shape: ViewportShape
    pixelHeight: int
    pixelWidth: int
    currentPixelHeight: int
    currentPixelWidth: int
    dpi: int
    touch: List[str] = None # TODO: model (no docs!)
    keyboard: List[str] = None # TODO: model
    video: dict = None # TODO: model

# -----

@dataclass
class FulfillmentContext:
    # "Alexa.Presentation.APL": ... ☠**☢ϟϟ☠☠☣
    System: FulfillmentContextSystem
    AudioPlayer: FulfillmentContextAudioPlayer = None
    Viewport: FulfillmentContextViewport = None
    Viewports: List[dict] = None # TODO: model

#
# Request
#

class RequestType(Enum):
    LAUNCH = "LaunchRequest"
    INTENT = "IntentRequest"
    SESSION_ENDED = "SessionEndedRequest"
    CAN_FULFILL_INTENT = "CanFulfillIntentRequest"

@dataclass
class FulfillmentRequest:
    type: RequestType
    requestId: str
    timestamp: datetime # TODO: timezone? This will be checked to prevent reply attacks
    locale: str

@dataclass
class FulfillmentLaunchRequest(FulfillmentRequest):
    pass

@dataclass
class FulfillmentIntentRequestIntent:
    name: str
    confirmationStatus: IntentConfirmationStatus
    slots: Dict[str, IntentSlot] = field(default_factory=dict)

@dataclass
class FulfillmentIntentRequest(FulfillmentRequest):
    intent: FulfillmentIntentRequestIntent
    dialogState: str = None

class SessionEndedReason(Enum):
    USER_INITIATED = "USER_INITIATED"
    ERROR = "ERROR"
    EXCEEDED_MAX_REPROMPTS = "EXCEEDED_MAX_REPROMPTS"

class SessionEndedErrorType(Enum):
    INVALID_RESPONSE = "INVALID_RESPONSE"
    DEVICE_COMMUNICATION_ERROR = "DEVICE_COMMUNICATION_ERROR"
    INTERNAL_SERVICE_ERROR = "INTERNAL_SERVICE_ERROR"
    ENDPOINT_TIMEOUT = "ENDPOINT_TIMEOUT"

@dataclass
class FulfillmentSessionEndedRequestError:
    type: SessionEndedErrorType
    message: str

@dataclass
class FulfillmentSessionEndedRequest(FulfillmentRequest):
    reason: SessionEndedReason
    error: FulfillmentSessionEndedRequestError = None

class CanFulfill(Enum):
    YES = "YES"
    NO = "NO"
    MAYBE = "MAYBE"

@dataclass
class FulfillmentCanFulfillIntentRequest:
    canFulfill: CanFulfill
    slots: dict # TODO: model

#
# Main request body
#

@dataclass
class FulfillmentBody:
    version: str
    session: FulfillmentSession
    context: FulfillmentContext
    request: Union[
        FulfillmentCanFulfillIntentRequest,
        FulfillmentSessionEndedRequest,
        FulfillmentIntentRequest,
        FulfillmentLaunchRequest
    ]

def from_dict(data: dict, data_class: type=FulfillmentBody):
    """
    Wraps :func:`dacite.from_dict` configuring enums for Alexa Fulfillment requests

    Args:
        data_class: The dataclass to use as schema
        data: Will be converted into a dataclass instance
    """    
    if data_class is FulfillmentBody:
        request_type = RequestType(data["request"]["type"])
       
        if request_type == RequestType.INTENT:
            request = from_dict(data["request"], data_class=FulfillmentIntentRequest)
        elif request_type == RequestType.LAUNCH:
            request = from_dict(data["request"], data_class=FulfillmentLaunchRequest)
        elif request_type == RequestType.SESSION_ENDED:
            request = from_dict(data["request"], data_class=FulfillmentSessionEndedRequest)
        elif request_type == RequestType.CAN_FULFILL_INTENT:
            request = from_dict(data["request"], data_class=FulfillmentCanFulfillIntentRequest)
            
        return FulfillmentBody(
            version=data["version"],
            session=from_dict(data["session"], data_class=FulfillmentSession),
            context=from_dict(data["context"], data_class=FulfillmentContext),
            request=request
        )

    return dacite.from_dict(
        data_class=data_class,
        data=data,
        config=dacite.Config(
            cast=[
                PlayerActivity,
                ViewportMode,
                ViewportShape,
                IntentConfirmationStatus,
                SlotType,
                SessionEndedReason,
                SessionEndedErrorType,
                RequestType
            ],
            type_hooks={
                datetime: lambda x: datetime.strptime(x, '%Y-%m-%dT%H:%M:%SZ')
            }
        )
    )


import json

class jsondict(dict):
    """
    A dictionary that is serialized to JSON when casted to `str`. This is
    convenient in logs.
    """

    def __str__(self):
        return json.dumps(self, indent=2)

    def __repr__(self):
        return self.__str__()



from datetime import datetime

import pytest

from intents import Sys
from intents.connectors._experimental.snips import entities
from intents.connectors._experimental.snips import prediction_format as pf

def test_date_mapping_from_service():
    mapping = entities.DateMapping()
    snips_date_result = {
        'input': 'My birthday is on august 24',
        'intent': {
            'intentName': 'UserSaysBirthday',
            'probability': 1.0
        },
        'slots': [
            {
                'range': {'start': 15, 'end': 27},
                'rawValue': 'on august 24',
                'value': {
                    'kind': 'InstantTime',
                    'value': '2021-08-24 00:00:00 +02:00',
                    'grain': 'Day',
                    'precision': 'Exact'
                },
                'entity': 'snips/date',
                'slotName': 'birthday_date'
            }
        ]
    }
    parse_result = pf.from_dict(snips_date_result)
    entity = mapping.from_service(parse_result.slots[0].value)
    assert entity == Sys.Date(2021, 8, 24)

def test_date_mapping_unexpected_grain():
    mapping = entities.DateMapping()
    value = {
        'kind': 'InstantTime',
        'value': '2021-08-24 00:00:00 +02:00',
        'grain': 'Month',
        'precision': 'Exact'
    }
    entity = mapping.from_service(value)
    with pytest.warns(None):
        assert entity == Sys.Date(2021, 8, 24)

def test_date_mapping_unexpected_kind():
    mapping = entities.DateMapping()
    value = {
        'kind': 'UNEXPECTED',
        'value': '2021-08-24 00:00:00 +02:00',
        'grain': 'Day',
        'precision': 'Exact'
    }
    entity = mapping.from_service(value)
    with pytest.warns(None):
        assert entity == Sys.Date(2021, 8, 24)

def test_date_mapping_to_service():
    mapping = entities.DateMapping()
    assert mapping.to_service(Sys.Date(2021, 8, 8)) == "2021-08-08"
    assert mapping.to_service(datetime(year=2021, month=8, day=8)) == "2021-08-08"



"""
An official Python implementation exists, which is more comprehensive, as well as more
complex: https://github.com/alexa/alexa-apis-for-python/blob/master/ask-smapi-model/ask_smapi_model/v1/skill/interaction_model/language_model.py
"""
from enum import Enum
from typing import List, Union
from dataclasses import dataclass, field

from intents.helpers.data_classes import OmitNone
from intents.connectors._experimental.alexa.slot_types import SystemSlotTypes

#
# Language Model
#

@dataclass
class LanguageModelIntentSlotMultipleValues:
    enabled: bool

@dataclass
class LanguageModelIntentSlot:
    name: str
    type: Union[SystemSlotTypes, str]
    samples: List[str] = OmitNone() # TODO: check
    multipleValues: LanguageModelIntentSlotMultipleValues = OmitNone()

@dataclass
class LanguageModelIntent:
    name: str
    slots: List[LanguageModelIntentSlot] = field(default_factory=list)
    samples: List[str] = field(default_factory=list)

@dataclass
class LanguageModelTypeValueName:
    value: str
    synonyms: List[str] = OmitNone()

@dataclass
class LanguageModelTypeValue:
    id: str
    name: LanguageModelTypeValueName

@dataclass
class LanguageModelType:
    name: str
    values: List[LanguageModelTypeValue]
    # TODO: valueSupplier

class FallbackIntentSensitivity(Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"

@dataclass
class LanguageModelConfigurationFallbackSensitivity:
    level: FallbackIntentSensitivity = FallbackIntentSensitivity.LOW

@dataclass
class LanguageModelConfiguration:
    fallbackIntentSensitivity: LanguageModelConfigurationFallbackSensitivity = OmitNone()

@dataclass
class LanguageModel:
    invocationName: str
    intents: List[LanguageModelIntent]
    types: List[LanguageModelType] = field(default_factory=list)
    modelConfiguration: LanguageModelConfiguration = OmitNone()

#
# Dialog
#

@dataclass
class Dialog:
    pass

#
# Prompts
#

@dataclass
class Prompt:
    pass

#
# Root
#

@dataclass
class InteractionModel:
    languageModel: LanguageModel
    dialog: Dialog = OmitNone()
    prompts: List[Prompt] = OmitNone()

@dataclass
class Agent:
    interactionModel: InteractionModel
    # TODO: complete



import pytest

from intents import Entity

def test_entity_default_name():

    class MyEntity(Entity):
        pass

    assert MyEntity.name == 'MyEntity'

def test_entity_custom_name():

    class MyEntity(Entity):
        name = "CustomEntityName"

    assert MyEntity.name == 'CustomEntityName'

def test_entity_reserved_name():
    with pytest.raises(ValueError):
        class I_MyEntity(Entity):
            pass

    with pytest.raises(ValueError):
        class MyEntity(Entity):
            name = "i_my_entity"



"""
Each of your Entity classes is supposed to define language data in a
`language/<LANGUAGE-CODE>/ENTITY_MyEntityClass.yaml` YAML file.

An Entity language file simply contains a list of entries and their synonyms.
For instance, this is the content of `example_agent/language/en/ENTITY_PizzaType.yaml`:

.. code-block:: yaml

    entries:
      Margherita:
        - normal
        - standard
      Diavola:
        - spicy
        - pepperoni
      Capricciosa:
        - with olives and artichokes

Check out :mod:`example_agent.restaurant` for a hands-on example with custom entities.
"""
import os
from dataclasses import dataclass
from typing import List, Dict, Union

import yaml

# pylint: disable=unused-import
import intents # (needed for building docs)
from intents import LanguageCode
from intents.model.entity import EntityType
from intents.language.agent_language import agent_language_folder

@dataclass
class EntityEntry:
    """
    Model Language entry for a Custom Entity. `EntityEntry` objects are produced
    by :func:`entity_language_data` as a result of parsing YAML language
    resources.
    
    Args:
        value: The canonical value of the entry (e.g. "Diavola")
        synonyms: A set of synonyms that refer to the same entry (e.g. "spicy",
            "pepperoni", ...)
    """
    value: str
    synonyms: List[str]

def make_language_data(entries: List[Union[str, List[str]]]) -> List[EntityEntry]:
    """
    Synthesize entity language data for an Entity from a list of entries. Each
    entry can just be a string, or a list where the first element is the value,
    and the remaining ones are synonyms.

    This function is mostly for internal use: it is recommended that you store
    language data in YAML files instead.

    Args:
        entries: A list of entity entries to build language data
    """
    result = []
    for e in entries:
        assert e
        if isinstance(e, str):
            e = [e]
        result.append(EntityEntry(value=e[0], synonyms=e[1:]))
    return result

def entity_language_data(
    agent_cls: "intents.model.agent.AgentType",
    entity_cls: EntityType,
    language_code: LanguageCode=None
) -> Dict[LanguageCode, List[EntityEntry]]:
    """
    Return language data for a given Entity.

    Args:
        agent_cls: The Agent class that registered the Entity
        entity_cls: The Entity class to load language resources for
        language_code: A specific Language to load. If not present, all
            available languages will be returned
    """
    # Custom language data
    if entity_cls.__entity_language_data__:
        # TODO: validate custom language data
        if language_code:
            return {language_code: entity_cls.__entity_language_data__[language_code]}
        return entity_cls.__entity_language_data__
    
    language_folder = agent_language_folder(agent_cls)

    if not language_code:
        result = {}
        for language_code in agent_cls.languages:
            language_data = entity_language_data(agent_cls, entity_cls, language_code)
            result[language_code] = language_data[language_code]
        return result

    language_file = os.path.join(language_folder, language_code.value, f"ENTITY_{entity_cls.name}.yaml")
    if not os.path.isfile(language_file):
        raise ValueError(f"Language file not found for entity '{entity_cls.name}'. Expected path: {language_file}.")

    with open(language_file, 'r') as f:
        language_data = yaml.load(f.read(), Loader=yaml.FullLoader)

    if not language_data:
        return []

    entries_data = language_data.get('entries', [])
    if not isinstance(entries_data, dict):
        raise ValueError(f"Invalid Entity language data for entity {entity_cls.name}. Entries data must be a dict. Entries data: {entries_data}")

    entries = []
    for value, synonyms in entries_data.items():
        if not isinstance(synonyms, list):
            raise ValueError(f"Invalid language data for entry {entity_cls.name}. Synonims data must always be lists. Synonims data for '{value}': '{synonyms}'")
        entries.append(EntityEntry(value, synonyms))

    return {language_code: entries}



from typing import Type
from unittest.mock import patch
from dataclasses import dataclass

import pytest

from intents import Intent, Entity, Agent, LanguageCode
from intents.language import EntityEntry
from intents.connectors._experimental.alexa import language

def _get_toy_agent() -> Type[Agent]:
    class ToyAgent(Agent):
        languages = ['en', 'it']
    return ToyAgent

class ToyEntity(Entity):
    __entity_language_data__ = {
        LanguageCode.ENGLISH: [
            EntityEntry("*", ["multiplication", "asterisk"]), # invalid, but synonyms
            EntityEntry("plus", ['+']),
            EntityEntry("$$$", ['cash', 'money']),
        ],
        LanguageCode.ITALIAN: [
            EntityEntry("*", ["moltiplicazione", "per"]), # invalid, but synonyms
            EntityEntry("addizione", ['+', "più"]), # TODO: canonical value = "più"
            EntityEntry("$$$", ['cash', 'soldi']),
        ]
    }

@dataclass
class ToyIntent(Intent):
    foo: ToyEntity

class ToyInvalidEntity(Entity):
    __entity_language_data__ = {
        LanguageCode.ENGLISH: [
            EntityEntry("this", ['is', 'ok']),
            EntityEntry("not ok: *", []),
            EntityEntry("this", ['also', 'good']),
        ],
        LanguageCode.ITALIAN: [
            EntityEntry("questa", ['è', 'ok']),
            EntityEntry("non ok: *", []),
            EntityEntry("questa", ['anche', 'bene']),
        ]
    }

@dataclass
class ToyInvalidIntent(Intent):
    bar: ToyInvalidEntity

@patch('intents.language.intent_language_data')
def test_language_data_only_one_language(*args):
    ToyAgent = _get_toy_agent()
    ToyAgent.register(ToyIntent)
    lc = language.AlexaLanguageComponent(ToyAgent)
    language_data = lc.entity_language_data(ToyEntity, LanguageCode.ITALIAN)
    assert list(language_data.keys()) == [LanguageCode.ITALIAN]

@patch('intents.language.intent_language_data')
def test_language_data_removes_invalid(*args):
    ToyAgent = _get_toy_agent()
    ToyAgent.register(ToyIntent)
    lc = language.AlexaLanguageComponent(ToyAgent)
    language_data = lc.entity_language_data(ToyEntity)

    expected_english = [
        language.AlexaEntityEntry(
            value="*", synonyms=["multiplication", "asterisk"],
            alexa_value="multiplication", alexa_synonyms=["asterisk"]
        ),
        language.AlexaEntityEntry(
            value="plus", synonyms=["+"],
            alexa_value="plus", alexa_synonyms=[]
        ),
        language.AlexaEntityEntry(
            value="$$$", synonyms=["cash", "money"],
            alexa_value="cash", alexa_synonyms=["money"]
        ),
    ]
    assert language_data[LanguageCode.ENGLISH] == expected_english

    expected_italian = [
        language.AlexaEntityEntry(
            value="*", synonyms=["moltiplicazione", "per"],
            alexa_value="moltiplicazione", alexa_synonyms=["per"]
        ),
        language.AlexaEntityEntry(
            value="addizione", synonyms=['+', "più"],
            alexa_value="addizione", alexa_synonyms=[]
        ),
        language.AlexaEntityEntry(
            value="$$$", synonyms=["cash", "soldi"],
            alexa_value="cash", alexa_synonyms=["soldi"]
        ),
    ]
    assert language_data[LanguageCode.ITALIAN] == expected_italian

@patch('intents.language.intent_language_data')
def test_id_to_value(*args):
    ToyAgent = _get_toy_agent()
    ToyAgent.register(ToyIntent)

    lc = language.AlexaLanguageComponent(ToyAgent)

    assert len(lc._entry_id_to_value[LanguageCode.ENGLISH]) == len(ToyEntity.__entity_language_data__[LanguageCode.ENGLISH])
    assert len(lc._entry_id_to_value[LanguageCode.ITALIAN]) == len(ToyEntity.__entity_language_data__[LanguageCode.ITALIAN])
    assert lc.alexa_entry_id_to_value("ToyEntity-multiplication", LanguageCode.ENGLISH) == "*"
    assert lc.alexa_entry_id_to_value("ToyEntity-moltiplicazione", LanguageCode.ITALIAN) == "*"

@patch('intents.language.intent_language_data')
def invalid_entities_raise_error(*args):
    ToyAgent = _get_toy_agent()
    ToyAgent.register(ToyInvalidIntent)

    with pytest.raises(ValueError):
        language.AlexaLanguageComponent(ToyAgent)



"""
Features:

- [OK] intent
- [OK] example utterances
- [OK] example utterances with parameter references
- [OK] system entity slots
- [OK] custom entity slots
- [  ] plain text responses
- [  ] intent relations
- [OK] multi language

Note that there are some relevant TODOs in this module (e.g. intent names may be
ambiguous, punctuation is stripped from utterances, and so on..)
"""
import re
import logging
from typing import List, Dict, Type
from dataclasses import asdict

from intents import Intent, Agent, EntityMixin
from intents.language import intent_language, intent_language_data, agent_supported_languages, LanguageCode
from intents.connectors._experimental.alexa import agent_schemas as ask_schema
from intents.connectors._experimental.alexa import names, language

from intents.helpers.data_classes import custom_asdict_factory

logger = logging.getLogger(__name__)

# TODO: model in framework
DEFAULT_INTENTS = [
    ask_schema.LanguageModelIntent(
        name="AMAZON.CancelIntent",
        samples=[]
    ),
    ask_schema.LanguageModelIntent(
        name="AMAZON.HelpIntent",
        samples=[]
    ),
    ask_schema.LanguageModelIntent(
        name="AMAZON.StopIntent",
        samples=[]
    ),
    ask_schema.LanguageModelIntent(
        name="AMAZON.NavigateHomeIntent",
        samples=[]
    ),
    ask_schema.LanguageModelIntent(
        name="AMAZON.FallbackIntent",
        samples=[]
    )
]

class AlexaExportComponent:
    agent_cls: Type[Agent]
    names_component: names.AlexaNamesComponent
    language_component: language.AlexaLanguageComponent
    invocation_name: str

    def __init__(self,
        agent_cls: Type[Agent],
        names_component: names.AlexaNamesComponent,
        language_component: language.AlexaLanguageComponent,
        invocation_name: str
    ):
        self.agent_cls = agent_cls
        self.names_component = names_component
        self.language_component = language_component
        self.invocation_name = invocation_name

    def render(self) -> Dict[LanguageCode, dict]:
        languages: List[LanguageCode] = agent_supported_languages(self.agent_cls)
        result = {}
        for lang in languages:
            rendered = self.render_agent(lang)
            rendered.interactionModel.languageModel.intents.extend(DEFAULT_INTENTS)
            result[lang] = asdict(rendered, dict_factory=custom_asdict_factory())
        return result

    def render_agent(self, lang: LanguageCode) -> ask_schema.Agent:
        return ask_schema.Agent(
            interactionModel=self.render_interaction_model(lang)
        )

    def render_interaction_model(self, lang: LanguageCode) -> ask_schema.InteractionModel:
        return ask_schema.InteractionModel(
            languageModel=self.render_language_model(lang)
        )

    def render_language_model(self, lang: LanguageCode) -> ask_schema.LanguageModel:
        intents = [self.render_intent(i, lang) for i in self.agent_cls.intents]
        intents = [i for i in intents if i]
        return ask_schema.LanguageModel(
            invocationName=self.invocation_name,
            intents=intents,
            types=[self.render_slot_type(e, lang) for e in self.agent_cls._entities_by_name.values()]
            # TODO: complete
        )

    #
    # Intent
    #

    def render_intent(self, intent_cls: Type[Intent], lang: LanguageCode) -> ask_schema.LanguageModelIntent:
        """
        Return None if intent has no example utterances
        """
        slots = self.render_intent_slots(intent_cls)
        samples = self.render_intent_samples(intent_cls, lang)
        if not samples:
            return None
        return ask_schema.LanguageModelIntent(
            name=self.names_component.intent_to_alexa(intent_cls),
            slots=slots,
            samples=samples
            # TODO: complete
        )

    def render_intent_slots(self, intent_cls: Type[Intent]) -> List[ask_schema.LanguageModelIntentSlot]:
        result = []
        for param_name, param_metadata in intent_cls.parameter_schema.items():
            entity_cls = param_metadata.entity_cls
            slot_type = self.names_component.entity_service_name(entity_cls)

            result.append(ask_schema.LanguageModelIntentSlot(
                name=param_name,
                type=slot_type,
                multipleValues=ask_schema.LanguageModelIntentSlotMultipleValues(
                    enabled=param_metadata.is_list
                )
            ))
        return result


    def render_intent_samples(self, intent_cls: Type[Intent], lang: LanguageCode) -> List[str]:
        language_data = intent_language_data(self.agent_cls, intent_cls, lang)
        language_data = language_data[lang]
        result = []
        for utterance in language_data.example_utterances:
            rendered_chunks = []
            for chunk in utterance.chunks():
                if isinstance(chunk, intent_language.TextUtteranceChunk):
                    rendered_chunks.append(chunk.text)
                elif isinstance(chunk, intent_language.EntityUtteranceChunk):
                    rendered_chunks.append("{" + chunk.parameter_name + "}")
                else:
                    raise ValueError(f"Unsupported utterance chunk type {type(chunk)}. This looks like a bug, please file an issue at https://github.com/dariowho/intents")
            utterance = "".join(rendered_chunks)
            # TODO: refine, especially for List parameters. Also, "{", "}" and
            # "_" are only allowed in slot references
            utterance = re.sub(r"[^a-zA-Z0-9 \-\{\}\_\.\']+", '', utterance)
            result.append(utterance)
        return result

    #
    # Slot Type
    #

    def render_slot_type(self, entity_cls: Type[EntityMixin], lang: LanguageCode) -> ask_schema.LanguageModelType:
        language_data = self.language_component.entity_language_data(entity_cls, lang)
        language_data = language_data[lang]
        slot_values = [self.render_slot_value(entity_cls, entry) for entry in language_data]
        slot_values = [v for v in slot_values if v]
        return ask_schema.LanguageModelType(
            name=entity_cls.name,
            values=slot_values
        )

    def render_slot_value(self, entity_cls: Type[EntityMixin], entity_entry: language.AlexaEntityEntry) -> ask_schema.LanguageModelTypeValue:
        return ask_schema.LanguageModelTypeValue(
            id=self.language_component.entry_value_id(entity_cls, entity_entry),
            name=ask_schema.LanguageModelTypeValueName(
                value=entity_entry.alexa_value,
                synonyms=entity_entry.alexa_synonyms
            )
        )

# from example_agent.agent import ExampleAgent
# from intents.connectors._experimental.alexa import AlexaConnector

# alexa = AlexaConnector(ExampleAgent, "any invocation")
# # render(alexa)

# alexa.export("./TMP_ALEXA")


from intents import Sys
from intents.connectors._experimental.alexa.slot_types import ENTITY_MAPPINGS

def test_date_mapping():
    entity_type = Sys.Date
    entity_value = "2021-07-11"
    mapping = ENTITY_MAPPINGS[entity_type]
    assert mapping.from_service(entity_value) == Sys.Date(2021, 7, 11)
    assert mapping.to_service(Sys.Date(2021, 7, 11)) == entity_value



"""
Here we export a :class:`Agent` to Dialogflow format.
"""
import os
import json
import shutil
import logging
import tempfile
from uuid import uuid1
from dataclasses import asdict
from typing import List, Dict, Set, Iterable, Type

from intents import Intent, EntityMixin, language
from intents.model.relations import intent_relations, FollowIntentRelation
import intents.connectors.dialogflow_es.agent_format as df
import intents.connectors.dialogflow_es.names as df_names
from intents.connectors.dialogflow_es.entities import MAPPINGS as ENTITY_MAPPINGS

logger = logging.getLogger(__name__)

def export(connector: "intents.DialogflowEsConnector", output_path: str, agent_name="py-agent") -> None:
    """
    Export the given agent to the given path
    """
    agent_cls = connector.agent_cls

    output_dir = os.path.join(tempfile.gettempdir(), 'dialogflow-agents-export', agent_name)
    intents_dir = os.path.join(output_dir, 'intents')
    entities_dir = os.path.join(output_dir, 'entities')

    if os.path.isdir(intents_dir):
        logger.warning("Removing existing intents folder: %s", intents_dir)
        shutil.rmtree(intents_dir)

    if os.path.isdir(entities_dir):
        logger.warning("Removing existing entities folder: %s", entities_dir)
        shutil.rmtree(entities_dir)

    os.makedirs(intents_dir)
    os.makedirs(entities_dir)

    with open(os.path.join(output_dir, 'agent.json'), 'w') as f:
        languages = agent_cls.languages
        json.dump(asdict(render_agent(connector, agent_name, languages)), f, indent=2)

    with open(os.path.join(output_dir, 'package.json'), 'w') as f:
        json.dump({"version": "1.0.0"}, f, indent=2)

    for intent in agent_cls.intents:
        language_data = language.intent_language_data(agent_cls, intent)
        rendered_intent = render_intent(connector, intent, language_data)
        with open(os.path.join(intents_dir, f"{intent.name}.json"), "w") as f:
            json.dump(asdict(rendered_intent), f, indent=2)
        
        for language_code, language_code_data in language_data.items():
            rendered_intent_usersays = render_intent_usersays(agent_cls, intent, language_code, language_code_data.example_utterances)
            filename = f"{intent.name}_usersays_{language_code.value}.json"
            with open(os.path.join(intents_dir, filename), "w") as f:
                usersays_data = [asdict(x) for x in rendered_intent_usersays]
                json.dump(usersays_data, f, indent=2)

    for entity_cls in agent_cls._entities_by_name.values():
        language_data = language.entity_language_data(agent_cls, entity_cls)
        rendered_entity = render_entity(entity_cls)
        with open(os.path.join(entities_dir, f"{entity_cls.name}.json"), "w") as f:
            json.dump(asdict(rendered_entity), f, indent=2)

        for language_code, entries in language_data.items():
            rendered_entity_entries = render_entity_entries(agent_cls, entries)
            filename = f"{entity_cls.name}_entries_{language_code.value}.json"
            with open(os.path.join(entities_dir, filename), "w") as f:
                entries_data = [asdict(x) for x in rendered_entity_entries]
                json.dump(entries_data, f, indent=2)

    if output_path.endswith('.zip'):
        output_path = output_path[:-4]
    shutil.make_archive(output_path, 'zip', output_dir)

#
# Agent
#

def render_agent(connector: "intents.DialogflowEsConnector",  agent_name: str, languages: List[language.LanguageCode]):
    google_assistant = df.AgentGoogleAssistant(
        project=connector.gcp_project_id,
        oAuthLinking=df.AgentGoogleAssistantOauthLinking()
        # TODO: include Google Assistant configuration
    )

    if connector.webhook_configuration:
        webhook = df.AgentWebhook(
            available=True,
            url=connector.webhook_configuration.url,
            headers=connector.webhook_configuration.headers
        )
    else:
        webhook = df.AgentWebhook()

    languages = [l.value for l in languages]
    return df.Agent(
        displayName=agent_name,
        webhook=webhook,
        googleAssistant=google_assistant,
        language=languages[0],
        supportedLanguages=languages[1:]
    )

#
# Intent
#

def get_input_contexts(connector: "DialogflowEsConnector", intent_cls: Type[Intent]) -> List[str]:
    result = [df_names.context_name(r.target_cls) for r in intent_relations(intent_cls).follow]
    
    return result

class _AffectedContextsList(list):
    """
    A list that appends a context only if no other context with the same name
    is already in the list.
    """
    added_names: Set[str]

    def __init__(self):
        super().__init__()
        self.added_names = set()

    def append(self, x: df.AffectedContext):
        if x.name in self.added_names:
            return
        self.added_names.add(x.name)
        super().append(x)

    def extend(self, x_list: List[df.AffectedContext]):
        x_list = [x for x in x_list if x.name not in self.added_names]
        for x in x_list:
            self.added_names.add(x.name)
        super().extend(x_list)

def get_output_contexts(
    connector: "DialogflowEsConnector",
    intent_cls: Type[Intent],
    visited: List[Type[Intent]]=None
) -> List[df.AffectedContext]:
    """
    An Intent can output its own context (e.g. intent `OrderFish` can spawn
    context `c_orderfish`). However, this should only happen if that context is
    actually needed (i.e. if there are intents referencing it).

    Also, this procedure is recursive: if intent `OrderFish` inherits from
    intent `OrderItem`, and there are intents referencing `c_orderitem`, then
    `OrderFish` should output `c_orderitem`
    """
    if not visited:
        visited = []

    if intent_cls is Intent:
        return []

    result = _AffectedContextsList()

    # Spawn own context if needed (i.e. at least one other intent follows this one)
    if connector._intent_needs_context(intent_cls):
        name = df_names.context_name(intent_cls)
        result.append(df.AffectedContext(name, intent_cls.lifespan))
    
    # Re-spawn context of followed intents that need to re-define lifespan
    rel: FollowIntentRelation
    for rel in intent_relations(intent_cls).follow:
        if new_lifespan := rel.relation_parameters.new_lifespan:
            for related_cls in [rel.target_cls] + rel.target_cls.parent_intents():
                context_name = df_names.context_name(related_cls)
                result.append(df.AffectedContext(context_name, new_lifespan))

    visited.append(intent_cls)
    for super_cls in intent_cls.parent_intents():
        if super_cls not in visited and issubclass(super_cls, Intent):
            result.extend(get_output_contexts(connector, super_cls, visited))

    # We cast to list because of compatibility with `asdict()`
    return list(result)

def render_intent(connector: "DialogflowEsConnector", intent_cls: Type[Intent], language_data: Dict[language.LanguageCode, language.IntentLanguageData]):
    response = df.Response(
        affectedContexts=get_output_contexts(connector, intent_cls),
        parameters=render_parameters(intent_cls, language_data),
        messages=render_responses(intent_cls, language_data, connector.rich_platforms),
    )

    return df.Intent(
        id=str(uuid1()),
        name=intent_cls.name,
        contexts=get_input_contexts(connector, intent_cls),
        responses=[response],

        # TODO: re-enable
        # webhookUsed=intent_cls.metadata.intent_webhook_enabled,
        # webhookForSlotFilling=intent_cls.metadata.slot_filling_webhook_enabled,
        events=[df.Event(df_names.event_name(intent_cls))]
    )

def render_parameters(intent_cls: Type[Intent], language_data: Dict[language.LanguageCode, language.IntentLanguageData]):
    result = []
    for param_name, param_metadata in intent_cls.parameter_schema.items():
        entity_cls = param_metadata.entity_cls
        data_type = ENTITY_MAPPINGS.service_name(entity_cls)

        prompts = []
        for language_code, language_code_data in language_data.items():
            for prompt in language_code_data.slot_filling_prompts.get(param_name, []):
                prompts.append(df.Prompt(value=prompt, lang=language_code.value))

        result.append(df.Parameter(
            id=str(uuid1()),
            name=param_name,
            required=param_metadata.required,
            dataType=f'@{data_type}',
            value=f"${param_name}",
            defaultValue=param_metadata.default if not param_metadata.required else '',
            isList=param_metadata.is_list,
            prompts=prompts
        ))
    return result

def render_response(response: language.IntentResponse, language_code: language.LanguageCode, platform: str):
    """
    platform: None = "Default"
    """
    if isinstance(response, language.TextIntentResponse):
        response: language.TextIntentResponse
        return df.TextResponseMessage(
            lang=language_code.value,
            speech=response.choices,
            platform=platform
        )
    elif isinstance(response, language.QuickRepliesIntentResponse):
        response: language.QuickRepliesIntentResponse
        return df.QuickRepliesResponseMessage(
            lang=language_code.value,
            replies=response.replies,
            platform=platform
        )
    elif isinstance(response, language.ImageIntentResponse):
        response: language.ImageIntentResponse
        return df.ImageResponseMessage(
            lang=language_code.value,
            imageUrl=response.url,
            title=response.title if response.title else "",
            platform=platform
        )
    elif isinstance(response, language.CardIntentResponse):
        response: language.CardIntentResponse
        buttons = None
        if response.link:
            buttons = [df.CardResponseMessageButton(text="👁", postback=response.link)]
        return df.CardResponseMessage(
            lang=language_code.value,
            title=response.title,
            subtitle=response.subtitle,
            imageUrl=response.image,
            buttons=buttons,
            platform=platform
        )
    elif isinstance(response, language.CustomPayloadIntentResponse):
        response: language.CustomPayloadIntentResponse
        return df.CustomPayloadResponseMessage(
            lang=language_code.value,
            payload={
                response.name: response.payload
            },
            platform=platform
        )
    else:
        raise ValueError(f"Unsupported response type: {response}")

def render_responses(intent_cls: Type[Intent], language_data: Dict[language.LanguageCode, language.IntentLanguageData], rich_platforms: Iterable[str]):
    result = []

    for language_code, language_code_data in language_data.items():
        if not language_code_data.responses:
            result.append(df.ResponseMessage(language_code.value))
            continue

        for response_group, responses in language_code_data.responses.items():
            assert response_group in [language.IntentResponseGroup.DEFAULT, language.IntentResponseGroup.RICH]
            if response_group == language.IntentResponseGroup.RICH:
                platforms_to_render = rich_platforms
            else:
                platforms_to_render = (None,) # Dialogflow will put the response in "Default" when platform=None

            for res in responses:
                for platform in platforms_to_render:
                    result.append(render_response(res, language_code, platform))

    return result

def render_utterance_chunk(chunk: language.UtteranceChunk):
    if isinstance(chunk, language.TextUtteranceChunk):
        return df.UsersaysTextChunk(text=chunk.text, userDefined=True)

    if isinstance(chunk, language.EntityUtteranceChunk):
        chunk: language.EntityUtteranceChunk
        meta = ENTITY_MAPPINGS.service_name(chunk.entity_cls)
        
        return df.UsersaysEntityChunk(
            text=chunk.parameter_value,
            alias=chunk.parameter_name,
            meta=f'@{meta}',
            userDefined=True
        )

    raise ValueError(f"Unsupported Utterance Chunk Type: {chunk}")

def render_intent_usersays(agent_cls: type, intent: Type[Intent], language_code: language.LanguageCode, examples: List[language.ExampleUtterance]):
    result = []
    for e in examples:
        result.append(df.IntentUsersays(
            id=str(uuid1()),
            lang=language_code.value,
            data=[render_utterance_chunk(c) for c in e.chunks()]
        ))
    return result

#
# Entity
#

def render_entity(entity_cls: Type[EntityMixin]) -> df.Entity:
    metadata = entity_cls.metadata
    return df.Entity(
        id=str(uuid1()),
        name=entity_cls.name,
        isRegexp=metadata.regex_entity,
        automatedExpansion=metadata.automated_expansion,
        allowFuzzyExtraction=metadata.fuzzy_matching
    )

def render_entity_entries(entity_cls: Type[EntityMixin], entries: List[language.EntityEntry]) -> List[df.EntityEntry]:
    result = []
    for e in entries:
        result.append(df.EntityEntry(
            value=e.value,
            synonyms=[e.value] + e.synonyms
        ))
    return result




import logging
import datetime
from enum import Enum
from intents import Sys
from intents.connectors.interface import EntityMapping, StringEntityMapping, ServiceEntityMappings
from intents.connectors._experimental.alexa.dates import parse_alexa_date

logger = logging.getLogger(__name__)

class SystemSlotTypes(Enum):
    Color = "AMAZON.Color"
    Genre = "AMAZON.Genre"
    Musician = "AMAZON.Musician"
    Number = "AMAZON.NUMBER"
    Language = "AMAZON.Language"
    Person = "AMAZON.Person"
    PhoneNumber = "AMAZON.PhoneNumber"

class DummyMapping(EntityMapping):
    service_name = None
    entity_cls = None

    def __init__(self, entity_cls, service_name="NOT_IMPLEMENTED"):
        self.entity_cls = entity_cls
        self.service_name = service_name
    def from_service(self, service_data):
        raise NotImplementedError(f"This Connector uses a Dummy Mapping for entity {self.entity_cls}.")
    def to_service(self, entity):
        raise NotImplementedError(f"This Connector uses a Dummy Mapping for entity {self.entity_cls}.")

class DateMapping(EntityMapping):
    entity_cls = Sys.Date
    service_name = "AMAZON.DATE"
    
    def from_service(self, service_data: str) ->  Sys.Date:
        if not isinstance(service_data, str):
            raise ValueError("Found non-string value for AMAZON.DATE entity. This is possibly a bug, please file an issue at https://github.com/dariowho/intents")

        date_from, date_to = parse_alexa_date(service_data)
        if date_to:
            # TODO: There could be an Alexa-only entity that maps to date intervals
            logger.warning("Alexa slot of type AMAZON.DATE contains an interval (e.g. 'this weekend', 'next winter'...). This case is not handled yet: the first day of the interval will be returned")

        return Sys.Date.from_py_date(date_from)

    def to_service(self, entity: datetime.date) -> str:
        return str(entity)

ENTITY_MAPPINGS = ServiceEntityMappings.from_list([
    StringEntityMapping(Sys.Person, SystemSlotTypes.Person.value),
    StringEntityMapping(Sys.Color, SystemSlotTypes.Color.value),
    DateMapping(),
    DummyMapping(Sys.Email), # TODO: implement
    StringEntityMapping(Sys.Integer, SystemSlotTypes.Number.value),
    StringEntityMapping(Sys.PhoneNumber, SystemSlotTypes.PhoneNumber.value),
    StringEntityMapping(Sys.Language, SystemSlotTypes.Language.value),
    DummyMapping(Sys.Url), # TODO: implement
    StringEntityMapping(Sys.MusicArtist, SystemSlotTypes.Musician.value),
    StringEntityMapping(Sys.MusicGenre, SystemSlotTypes.Genre.value)
])


"""
Here we model Dialogflow payloads that are involved in webhook fulfillment calls.
"""
from typing import List
from dataclasses import dataclass

from intents.helpers.data_classes import OmitNone
from intents.connectors.dialogflow_es import prediction_format as pf

#
# Request
#

# Webhook requests are very similar to prediction requests. Models are the same
# ones defined in :mod:`prediction_format.py`

#
# Response
#

@dataclass
class WebhookResponseFollowupEvent:
    name: str
    languageCode: str
    parameters: dict

@dataclass
class WebhookResponse:
    followupEventInput: WebhookResponseFollowupEvent
    fulfillmentMessages: List[pf.QueryResultMessage] = OmitNone()
    # payload: TODO
    # outputContexts: TODO
    # sessionEntityTypes: TODO



"""
Entity Mappings are used to (de-)serialize entity types and values in the
communication between a
:class:`~intents.connectors.interface.connector.Connector` and its Service. This
is done when

* Connector exports Agent to Service; here system entity names (e.g.
  :class:`Sys.Integer`) must be converted to Service name (e.g.
  `@sys.number-integer` in Dialogflow)
* Connector processes a Prediction or Fulfillment payload; in this case, not
  only a reverse lookup is needed, but it may be necessary to de-serialize an
  entity value as well. For innstance, ISO datetime strings mut be cast to
  Python datetime objects
* Connector triggers an Intent on Service; here parameter names must be
  converted, and values must be serialized

The base class for mappings is :class:`EntityMapping`. Two builtin convenience
mappings are also available: :class:`StringEntityMapping`,
and :class:`PatchedEntityMapping`.

It is convenient to collect Connector mappings in a
:class:`ServiceEntityMapping` dict, which implements some utility methods
related to mapping collections.

Finally, here we also define :func:`deserialize_intent_parameters`, a helper
that turns a dict of parameters in the Service format and converts it into a
dict of *Intents* entities.
"""

import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, List, Dict, Type

from intents import Intent, Entity, LanguageCode
from intents.model.entity import EntityMixin, SystemEntityMixin

logger = logging.getLogger(__name__)

# TODO: turn it to an abstract class, when pylint will support dataclass
# implementation of abstract properties
class EntityMapping():
    """
    An Entity Mapping is a (de-)serializer for predicted Entities.

    Most of the times a Mapping is not needed as the Entity can be mapped
    directly to its type (e.g. "3" -> `Number(3)`). However, prediction services
    such as Dialogflow may define system entities of structured types; a notable
    example is Dialogflow's *sys.person* entity, which is returned as `{"name":
    "John"}` and therefore needs custom logic to be mapped to `Person("John")`.
    This is modelled in :class:`intents.connectors.dialogflow_es.entities.PersonEntityMapping`.

    Another notable scenario is Date/Time objects. A mapping can be used to
    convert time strings from the Service format to python objects. For
    Dialogflow ES, this is modelled in  :class:`intents.connectors.dialogflow_es.entities.DateTimeEntityMapping`.
    """

    @property
    def entity_cls(self) -> Type[EntityMixin]:
        """
        This is the internal entity type that is being mapped.

        >>> mapping = StringEntityMapping(Sys.Integer, 'sys.number-integer')
        >>> mapping.entity_cls
        Sys.Integer
        """
        raise NotImplementedError()

    @property
    def service_name(self) -> str:
        """
        This is the name of the Entity in the Prediction Service domain.

        >>> mapping = StringEntityMapping(Sys.Integer, 'sys.number-integer')
        >>> mapping.service_name
        'sys.number-integer'
        """
        raise NotImplementedError()

    @property
    def supported_languages(self) -> List[LanguageCode]:
        """
        It may happen that a Prediction Service only supports an Entity for a
        limited set of languages. For instance, Snips NLU only supports its
        `snips/date` entity in English.

        This property is set to `None` when the mapping is valid for all the
        service-supported language. If support is restricted to a subset of
        them, it will contain the list of language codes. Export and prediction
        procedures must handle accordingly.
        """
        return None

    @abstractmethod
    def from_service(self, service_data: Any) -> SystemEntityMixin:
        """
        De-serialize the Service representation of an Entity (typically the value
        that is returned at prediction time) to an instance of one of the internal Entity
        classes in :class:`intents.model.entities`

        >>> date_mapping.from_service("2021-07-11")
        Sys.Date(2021, 7, 11)

        :param service_data: A parameter value, as it is returned by the Service
                             in a prediction/trigger response
        :return: the parameter value, modelled as one of the System Entity classes
        """

    @abstractmethod
    def to_service(self, entity: SystemEntityMixin) -> Any:
        """
        Serialize a System Entity instance into a Service representation (typically,
        to be sent as a parameter of a trigger request)

        >>> date_mapping.to_service(Sys.Date(2021, 7, 11))
        "2021-07-11"

        Args:
            entity: The System Entity to serialize

        Returns:
            The serialized Entity that can be sent to Service (e.g. in a trigger request)
        """

@dataclass(frozen=True)
class StringEntityMapping(EntityMapping):
    """
    This is a generic :class:`EntityMapping` that reads values as they are sent
    by the prediction service (e.g. `"3"` -> `Sys.Integer("3")`), and serializes
    by simple string conversion (e.g. `Sys.Integer(3)` -> "3"). This is the most
    common case when dealing with entities.

    The System Entity to use must be defined when instantiating the mapping, for
    instance:

    >>> StringEntityMapping(Sys.Integer, "sys.number-integer")

    Args:
        entity_cls: One of the `Sys.*` entity classes
        service_name: Name of the corresponding entity within the Prediction
            Service
    """

    entity_cls: Type[EntityMixin] = None
    service_name: str = None

    def from_service(self, service_data: Any) -> SystemEntityMixin:
        return self.entity_cls(service_data)

    def to_service(self, entity: SystemEntityMixin) -> Any:
        return str(entity)

@dataclass(frozen=True)
class PatchedEntityMapping(EntityMapping):
    """
    Different Prediction Services support different entities. For instance,
    :class:`Sys.Color` is native in Dialogflow, but is not supported in Snips.
    In some cases, we can patch missing system entities with custom ones; for
    instance, :class:`Sys.Color` can be patched with builtin
    :class:`~intents.resources.builtin_entities.color.I_IntentsColor`.
    `PatchedEntityMapping` can be use to define mappings for system entities
    that are patched with custom ones.

    Connectors that use patched entities must define logic to handle
    `PatchedEntityMapping` in their export procedures.

    By default, `builtin_entity.name` is used as Service entity name, and values
    are (de)serialized as simple strings. If a Connector have different
    required, it should define a custom subclass of `PatchedEntityMapping`.
    """
    entity_cls: Type[EntityMixin] = None
    builtin_entity: Entity = None

    @property
    def service_name(self):
        return self.builtin_entity.name

    def from_service(self, service_data: Any) -> SystemEntityMixin:
        return self.entity_cls(service_data)

    def to_service(self, entity: SystemEntityMixin) -> Any:
        return str(entity)

class ServiceEntityMappings(dict):
    """
    Models a collection of entity mappings, in the form of a dict where the key is a
    System entity class (i.e. inherits from :class:`SystemEntityMixin`) and the value
    is a :class:`EntityMapping`. In addition to a standard dict, these features
    are added:

    * Instantiate from a list of mappings with :meth:`from_list`
    * Flexible lookup with :meth:`ServiceEntityMapping.lookup`
    """

    def lookup(self, entity_cls: Type[EntityMixin]) -> EntityMapping:
        """
        Return the mapping in the dictionary that is associated with the given
        `entity_cls`. In addition to a simple `mappings[entity_cls]`, this
        method also implements a fallback for Custom Entities. That is, when a
        Custom Entity is not in the mapping dict (typically they are not),
        retrieve return an on-the-fly mapping generated with
        :meth:`~ServiceEntityMappings.custom_entity_mapping`.

        Args:
            entity_cls: The Entity class to lookup
        Returns:
            The mapping that refers to the given Entity class
        Raises:
            KeyError: If no mapping exists that can be used for `entity_cls`
        """
        # `Entity` objects are custom entities
        if issubclass(entity_cls, Entity) and entity_cls not in self:
            return self.custom_entity_mapping(entity_cls)
        if entity_cls not in self:
            mapped_entities = [m.entity_cls for m in self.values()]
            raise KeyError(f"Failed to lookup entity {entity_cls} in mappings. Mapped entities: {mapped_entities}")
        return self[entity_cls]

    def custom_entity_mapping(self, entity_cls: Type[EntityMixin]) -> EntityMapping:
        """
        Generate an entity mapping on the fly for the given custom entity. This
        is needed because, while System entities are static, Custom ones need to
        be handled dynamically at run time.

        By default, this returns a simple :class:`StringEntityMapping`, where
        `service_name` is `entity_cls.name`. Connectors may override this method
        to implement custom behavior.

        Args:
            entity_cls: A Custom Entity for which a mapping will be generated

        Returns:
            A mapping for the given Entity
        """
        return StringEntityMapping(
            entity_cls=entity_cls,
            service_name=entity_cls.name
        )

    def service_name(self, entity_cls: Type[EntityMixin]):
        """
        Return the name of the given entity in the specific service; this can be
        the class name itself, or an :class:`EntityMapping` lookup in the case
        of System Entities.

        For instance, a :class:`Sys.Person` Entity will need to be looked up in
        the mappings to find out its service name (`sys.person` in Dialogflow,
        `AMAZON.Person` in Alexa, and so on). A custom entity (e.g. `PizzaType`)
        will use its class name instead.
        """
        return self.lookup(entity_cls).service_name

    def is_mapped(self, entity_cls: Type[EntityMixin], lang: LanguageCode) -> bool:
        """
        Return `False` if no mapping is defined for the given entity. Also
        return `False` if a mapping exists, but the mapping defines
        :attr:`~EntityMapping.supported_languages` and the given language is not
        in the list.

        Args:
            entity_cls: The Entity class to lookup
            lang: The language that should be supported by the mapping
        """
        try:
            mapping = self.lookup(entity_cls)
        except KeyError:
            return False

        if not mapping.supported_languages:
            return True

        return lang in mapping.supported_languages

    @classmethod
    def from_list(cls, mapping_list: List[EntityMapping]) -> "ServiceEntityMappings":
        """
        Convenience method for building a mapping from a list, instead of
        specifying the whole dict. See Dialogflow's
        :mod:`~intents.connectors.dialogflow_es.entities` module for an example.

        Args:
            mapping_list: The mappings that will be used to build the
                `ServiceEntityMapping` dict
        Returns:
            A `ServiceEntityMapping` with the given mappings
        """
        result = cls()
        for mapping in mapping_list:
            if mapping in result:
                raise ValueError(f"Mapping {mapping} already defined in list: {mapping_list}")
            result[mapping.entity_cls] = mapping
        return result

def deserialize_intent_parameters(
    service_parameters: Dict[str, Any],
    intent_cls: Type[Intent],
    mappings: ServiceEntityMappings
) -> Dict[str, EntityMixin]:
    """
    Cast parameters from Service format to Intents format according to the given
    schema. Typically this happens when a Connector has to turn prediction
    parameters into *Intents* entities.

    Args:
        service_parameters: The parameters dict, as it is returned by a Prediction Service
        intent_cls: The Intent parameters will be matched against
        mappings: The Service Entity Mappings, to deserialize parameter values
    Return:
        A dictionary like `service_parameters`, but all the values are converted
        to native Intents Entity objects.
    """
    result = {}
    schema = intent_cls.parameter_schema
    for param_name, param_value in service_parameters.items():
        if param_name not in schema:
            raise ValueError(f"Found parameter {param_name} in Service Prediction, but Intent "
                             "class does not define it. Make sure your cloud Agent is in sync "
                             "with your local one. A new upload() may solve the issue")
        param_metadata = schema[param_name]
        entity_cls = param_metadata.entity_cls
        mapping = mappings.lookup(entity_cls)
        try:
            if param_metadata.is_list:
                if not isinstance(param_value, list):
                    raise ValueError(f"Parameter {param_name} is defined as List, but returned value is not of 'list' type: {param_value}")
                result[param_name] = [mapping.from_service(x) for x in param_value]
            else:
                result[param_name] = mapping.from_service(param_value)
        except Exception as exc:
            raise RuntimeError(f"Failed to match parameter '{param_name}' with value '{param_value}' against schema {schema}. See source exception above for details.") from exc

    return result



from google.cloud.dialogflow_v2.types import DetectIntentResponse
from google.cloud.dialogflow_v2 import types as df_types

from intents.language import IntentResponseGroup, TextIntentResponse, QuickRepliesIntentResponse, CardIntentResponse
from intents.connectors.dialogflow_es.prediction import DialogflowTextResponse, \
    DialogflowQuickRepliesResponse, \
    DialogflowCardResponse, \
    DetectIntentBody, \
    intent_responses, \
    build_response_message
from intents.connectors.dialogflow_es import prediction_format as df


# {'text': {'text': ['How about Hotel Belvedere, in Como?']}}
text_message_serialized = b'\n%\n#How about Hotel Belvedere, in Como?'
text_message = df_types.intent.Intent.Message.deserialize(text_message_serialized)
text_message_expected = TextIntentResponse(
    choices=['How about Hotel Belvedere, in Como?']
)

# {'text': {'text': ['I also like travels, how can I help you?']}, 'platform': 'TELEGRAM'}
text_message_rich_serialized = b'\n*\n(I also like travels, how can I help you?0\x03'
text_message_rich = df_types.intent.Intent.Message.deserialize(text_message_rich_serialized)
text_message_rich_expected = TextIntentResponse(
    choices=['I also like travels, how can I help you?']
)

# {'quickReplies': {'title': 'Quick Replies',
#   'quickReplies': ['Recommend an hotel', 'Send holiday photo']},
#  'platform': 'TELEGRAM'}
quick_replies_serialized = b'\x1a7\n\rQuick Replies\x12\x12Recommend an hotel\x12\x12Send holiday photo0\x03'
quick_replies_message = df_types.intent.Intent.Message.deserialize(quick_replies_serialized)
quick_replies_expected = QuickRepliesIntentResponse(
    replies=['Recommend an hotel', 'Send holiday photo']
)

# {'card': {'title': 'Hotel Belvedere',
#   'subtitle': '📍 Como, IT 💬 (42) 👍 ★★★★☆',
#   'imageUri': 'https://loremflickr.com/320/240/holiday',
#   'buttons': [{'text': '👁',
#     'postback': 'https://github.com/dariowho/intents'}]},
#  'platform': 'TELEGRAM'}
card_message_serialized = b'"\x95\x01\n\x0fHotel Belvedere\x12,\xf0\x9f\x93\x8d Como, IT \xf0\x9f\x92\xac (42) \xf0\x9f\x91\x8d \xe2\x98\x85\xe2\x98\x85\xe2\x98\x85\xe2\x98\x85\xe2\x98\x86\x1a\'https://loremflickr.com/320/240/holiday"+\n\x04\xf0\x9f\x91\x81\x12#https://github.com/dariowho/intents0\x03'
card_message = df_types.intent.Intent.Message.deserialize(card_message_serialized)
card_message_expected = CardIntentResponse(
    title='Hotel Belvedere',
    subtitle='📍 Como, IT 💬 (42) 👍 ★★★★☆',
    image='https://loremflickr.com/320/240/holiday',
    link='https://github.com/dariowho/intents'
)

# A full Dialogflow Response, with
# - A text message in the DEFAULT platform
# - A text message in the TELEGRAM platform
# - A quick replies message in the TELEGRAM platform
df_response_quick_replies_serialized = b'\n-1cedb9e6-f958-437f-9299-74f966fbec62-9779ea79\x12\xa2\x03\n\x10i want to travel"\x00(\x012QIf you like I can recommend you an hotel. Or I can send you some holiday pictures:U\nS\nQIf you like I can recommend you an hotel. Or I can send you some holiday pictures:.\n*\n(I also like travels, how can I help you?0\x03:;\x1a7\n\rQuick Replies\x12\x12Recommend an hotel\x12\x12Send holiday photo0\x03Zj\nOprojects/learning-dialogflow/agent/intents/e3a1e749-be67-11eb-8ad8-bbef97dc13e7\x12\x17travels.UserWantsTravele\x00\x00\x80?z\x02en'
df_response_quick_replies = DetectIntentResponse.deserialize(df_response_quick_replies_serialized)

def test_text_intent_response():
    message = df.from_protobuf(df.QueryResultMessage, text_message)
    result = DialogflowTextResponse.from_df_message(message)
    assert result == text_message_expected
    assert result == build_response_message(message)

    message = df.from_protobuf(df.QueryResultMessage, text_message_rich)
    result = DialogflowTextResponse.from_df_message(message)
    assert result == text_message_rich_expected
    assert result == build_response_message(message)

def test_quick_replies_response():
    message = df.from_protobuf(df.QueryResultMessage, quick_replies_message)
    result = DialogflowQuickRepliesResponse.from_df_message(message)
    assert result == quick_replies_expected
    assert result == build_response_message(message)

def test_card_response():
    message = df.from_protobuf(df.QueryResultMessage, card_message)
    result = DialogflowCardResponse.from_df_message(message)
    assert result == card_message_expected
    assert result == build_response_message(message)

def test_intent_responses():
    result = intent_responses(DetectIntentBody(df_response_quick_replies))
    expected = {
        IntentResponseGroup.DEFAULT: [
            TextIntentResponse(choices=["If you like I can recommend you an hotel. Or I can send you some holiday pictures"])
        ],
        IntentResponseGroup.RICH: [
            TextIntentResponse(choices=["I also like travels, how can I help you?"]),
            QuickRepliesIntentResponse(replies=["Recommend an hotel", "Send holiday photo"])
        ]
    }
    assert result == expected



import os
import tempfile
from unittest.mock import patch
from dataclasses import dataclass

from example_agent import ExampleAgent
from intents import Agent, Intent, follow
from intents.connectors.dialogflow_es import export, connector, entities, names
from intents.connectors.dialogflow_es import agent_format as df

#
# Testing intents
#

@dataclass
class FollowedIntent(Intent):
    pass

@dataclass
class FollowedIntentSubclass(FollowedIntent):
    pass

@dataclass
class FollowingIntent(Intent):
    parent_followed_intent: FollowedIntent = follow()

@dataclass
class FollowingIntentSubclass(FollowingIntent):
    pass

@dataclass
class FollowingSubclassIntent(Intent):
    parent_followed_intent_subclass: FollowedIntentSubclass = follow()

class MockAgent(Agent):
    pass

with patch('intents.language.intent_language_data'):
    MockAgent.register(FollowedIntent)
    MockAgent.register(FollowedIntentSubclass)
    MockAgent.register(FollowingIntent)
    MockAgent.register(FollowingIntentSubclass)
    MockAgent.register(FollowingSubclassIntent)

#
# Mock Connector
#

class MockDialogflowConnector(connector.DialogflowEsConnector):

    agent_cls: type = None
    _need_context_set = None


    def __init__(self, agent_cls):
        self.agent_cls = agent_cls
        self._need_context_set = connector._build_need_context_set(agent_cls)

    gcp_project_id: str = "fake-project-id"
    rich_platforms: tuple = ("telegram", "slack")
    webhook_configuration = None
    entity_mappings = entities.MAPPINGS

#
# Tests
#

def test_export_example_agent_no_exceptions():
    with tempfile.TemporaryDirectory() as temp_dir:
        export.export(MockDialogflowConnector(ExampleAgent), os.path.join(temp_dir, 'TMP_AGENT.zip'))

def test_get_input_contexts():
    mock_connector = MockDialogflowConnector(MockAgent)
    
    result = export.get_input_contexts(mock_connector, FollowedIntent)
    assert result == []

    result = export.get_input_contexts(mock_connector, FollowedIntentSubclass)
    assert result == []

    result = export.get_input_contexts(mock_connector, FollowingIntent)
    assert result == [names.context_name(FollowedIntent)]

    result = export.get_input_contexts(mock_connector, FollowingIntentSubclass)
    assert result == [names.context_name(FollowedIntent)]

    result = export.get_input_contexts(mock_connector, FollowingSubclassIntent)
    assert result == [names.context_name(FollowedIntentSubclass)]

def test_get_output_contexts():
    mock_connector = MockDialogflowConnector(MockAgent)
    _c = lambda x: names.context_name(x)

    result = export.get_output_contexts(mock_connector, FollowedIntent)
    assert result == [
        df.AffectedContext(name=_c(FollowedIntent), lifespan=5)
    ]

    result = export.get_output_contexts(mock_connector, FollowedIntentSubclass)
    assert result == [
        df.AffectedContext(name=_c(FollowedIntentSubclass), lifespan=5),
        df.AffectedContext(name=_c(FollowedIntent), lifespan=5),
    ]

    result = export.get_output_contexts(mock_connector, FollowingIntent)
    assert result == []

    result = export.get_output_contexts(mock_connector, FollowingIntentSubclass)
    assert result == []

    result = export.get_output_contexts(mock_connector, FollowingSubclassIntent)
    assert result == []

# TODO:
#   - own output context
#   - output context when re-defining lifespan of followed intent
#   - contexts do not appear twice
#   - superclasses are included in new_lifespan case



"""
This is the client-facing part of the interface. you will subclass the
:class:`Connector` base class, and clients will operate on its instances.
"""

import logging
from uuid import uuid1
from abc import ABC, abstractmethod
from typing import Union

from intents import Intent, Agent, LanguageCode
from intents.language import ensure_language_code, agent_supported_languages
from intents.connectors.interface.prediction import Prediction
from intents.connectors.interface.fulfillment import FulfillmentRequest

logger = logging.getLogger(__name__)

class Connector(ABC):
    """
    Connect the given Agent to a Prediction Service.

    Args:
        agent_cls: The Agent to connect
        default_session: A default session ID (conversation channel) for
            predictions. If None, Connector will generate a random string
        default_language: A default language for predictions. If None, Connector
            will use the Agent's firs defined language.
    """
    agent_cls: type(Agent)
    default_session: str
    default_language: str

    def __init__(
        self,
        agent_cls: type(Agent),
        default_session: str=None,
        default_language: Union[LanguageCode, str]=None
    ):
        if not default_language:
            default_language = agent_supported_languages(agent_cls)[0]
        default_language = ensure_language_code(default_language)
        
        if not default_session:
            default_session = f"py-{str(uuid1())}"
        
        self.agent_cls = agent_cls
        self.default_session = default_session
        self.default_language = default_language

    @abstractmethod
    def predict(self, message: str, session: str=None, language: Union[LanguageCode, str]=None) -> Prediction:
        """
        Predict the given User message in the given session using the given
        language. When `session` or `language` are None, `predict` will use the
        default values that are specified in :meth:`__init__`.

        *predict* will return an instance of :class:`Prediction`, representing
        the service response.

        >>> from intents.connectors import DialogflowEsConnector
        >>> from example_agent import ExampleAgent
        >>> df = DialogflowEsConnector('/path/to/service-account.json', ExampleAgent)
        >>> prediction = df.predict("Hi, my name is Guido")
        >>> prediction.intent
        UserNameGive(user_name='Guido')
        >>> prediction.intent.user_name
        "Guido"
        >>> prediction.fulfillment_text
        "Hi Guido, I'm Bot"
        >>> prediction.confidence
        0.86

        Args:
            message: The User message to predict
            session: Any string identifying a conversation
            language: A LanguageCode object, or a ISO 639-1 string (e.g. "en")
        """

    @abstractmethod
    def trigger(self, intent: Intent, session: str=None, language: Union[LanguageCode, str]=None) -> Prediction:
        """
        Trigger the given Intent in the given session using the given language.
        When `session` or `language` are None, `predict` will use the default
        values that are specified in :meth:`__init__`.

        >>> from intents.connectors import DialogflowEsConnector
        >>> from example_agent import ExampleAgent, smalltalk
        >>> df = DialogflowEsConnector('/path/to/service-account.json', ExampleAgent)
        >>> prediction = df.trigger(smalltalk.AgentNameGive(agent_name='Alice'))
        >>> prediction.intent
        AgentNameGive(agent_name='Alice')
        >>> prediction.fulfillment_text
        "Howdy Human, I'm Alice"
        >>> prediction.confidence
        1.0

        Args:
            intent: The Intent instance to trigger
            session: Any string identifying a conversation
            language: A LanguageCode object, or a ISO 639-1 string (e.g. "en")
        """

    @abstractmethod
    def fulfill(self, fulfillment_request: FulfillmentRequest) -> dict:
        """
        This method is responsible for handling requests coming from a
        fulfillment interface. We are at that point in the flow when an intent
        was triggered/predicted, and Service is calling the webhook service for
        fulfillment. Refer to :mod:`intents.fulfillment` for a more detailed
        explanation.

        In this method, Connector interprets the body of the request, builds a
        :class:`~intents.model.intent.FulfillmentContext` object, builds
        the :class:`~intents.model.Intent` object that is references in the
        request, and calls its :meth:`~intents.model.intent.Intent.fulfill`
        method.

        This will produce a
        :class:`~intents.model.intent.FulfillmentResult` object, that
        Connector will translate into a Service-compatible response (a
        dictionary) and return to caller.

        Args:
            fulfillment_request: A raw fulfillment request, as it is sent by the
                Prediction service (typically via a standard REST webhook call)
        Returns:
            An object containing a response that Service can read
        """

    @abstractmethod
    def upload(self):
        """
        Upload the connected Agent to the Prediction Service.
        """

    @abstractmethod
    def export(self, destination: str):
        """
        Export the connected Agent in a format that can be read and imported
        natively by the Prediction service. For instance, the Dialogflow service
        will produce a ZIP export that can be imported from the Dialogflow
        console.

        Note that you can also directly upload the Agent with :meth:`upload`.

        :param destination: destination path of the exported Agent
        """



from enum import Enum
from typing import Union

class LanguageCode(Enum):

    ENGLISH = 'en'
    ENGLISH_US = 'en_US'
    ENGLISH_UK = 'en_UK'
    ITALIAN = 'it'
    SPANISH = 'es'
    SPANISH_SPAIN = 'es_ES'
    SPANISH_LATIN_AMERICA = 'es_LA'
    GERMAN = 'de'
    FRENCH = 'fr'
    DUTCH = 'nl'
    CHINESE = 'zh'
    CHINESE_PRC = 'zh_CN'
    CHINESE_HONG_KONG = 'zh_HK'

LANGUAGE_CODES = [x.value for x in LanguageCode]

FALLBACK_LANGUAGE = {
    LanguageCode.ENGLISH: [
        LanguageCode.ENGLISH_UK,
        LanguageCode.ENGLISH_US
    ],
    LanguageCode.ENGLISH_US: [
        LanguageCode.ENGLISH,
        LanguageCode.ENGLISH_UK
    ],
    LanguageCode.ENGLISH_UK: [
        LanguageCode.ENGLISH,
        LanguageCode.ENGLISH_US
    ],
    LanguageCode.ITALIAN: [],
    LanguageCode.SPANISH: [
        LanguageCode.SPANISH_LATIN_AMERICA
    ],
    LanguageCode.SPANISH_SPAIN: [
        LanguageCode.SPANISH,
        LanguageCode.SPANISH_LATIN_AMERICA
    ],
    LanguageCode.SPANISH_LATIN_AMERICA: [
        LanguageCode.SPANISH,
        LanguageCode.SPANISH_SPAIN
    ],
    LanguageCode.GERMAN: [],
    LanguageCode.FRENCH: [],
    LanguageCode.DUTCH: [],
    LanguageCode.CHINESE: [
        LanguageCode.CHINESE_PRC
    ],
    LanguageCode.CHINESE_PRC: [
        LanguageCode.CHINESE
    ],
    LanguageCode.CHINESE_HONG_KONG: []
}

def ensure_language_code(lang: Union[LanguageCode, str]) -> LanguageCode:
    """
    Make sure `lang` is a :class:`LanguageCode` value. If not, return
    `LanguageCode(lang)` (assuming it's a string).

    This is useful for user-facing methods, where input may be a language code
    or a string.

    Args:
        lang: An input language code
    Returns:
        A LanguageCode object representing `lang` 
    """
    if isinstance(lang, LanguageCode):
        return lang
    return LanguageCode(lang)



import pytest
from unittest.mock import patch

from intents import LanguageCode
from intents.language.agent_language import match_agent_language

@patch("intents.language.agent_language.agent_supported_languages")
def test_match_agent_language__agent_has_language(m_supported_languages):
    m_supported_languages.return_value = [LanguageCode.ENGLISH, LanguageCode.ITALIAN]
    assert match_agent_language(None, LanguageCode.ITALIAN) == LanguageCode.ITALIAN

@patch("intents.language.agent_language.agent_supported_languages")
def test_match_agent_language__agent_has_fallback_language(m_supported_languages):
    m_supported_languages.return_value = [LanguageCode.ENGLISH, LanguageCode.ITALIAN]
    assert match_agent_language(None, LanguageCode.ENGLISH_US) == LanguageCode.ENGLISH

@patch("intents.language.agent_language.agent_supported_languages")
def test_match_agent_language__agent_has_no_language(m_supported_languages):
    m_supported_languages.return_value = [LanguageCode.ENGLISH, LanguageCode.ITALIAN]
    with pytest.raises(KeyError):
        match_agent_language(None, LanguageCode.CHINESE_HONG_KONG)



"""
Language resources are defined separately from :class:`Intent` and
:class:`Entity` classes. They are stored as plain YAML files in a `language/`
folder, aside Agent Python modules; this is to be flexible in the relationship
with designers and translators, as well as to allow some degree of automation
when downloading cloud changes back to the local Agent definition (this feature
is currently not implemented).

Your `language/` folder will contain one subfloder per language (i.e.
`language/en/`, `language/it/`, ...); each of these will contain

* one YAML file per Intent (e.g. `my_module.my_intent_class.yaml`). Intent
  language resources are specified in :mod:`intents.language.intent_language`
* one YAML file per entity (e.g. `ENTITY_MyEntityClass.yaml`). Entity language
  resources are specified in :mod:`intents.language.entity_language`

.. tip::

    It may be useful to look at The Example Agent code and language resources
    (https://github.com/dariowho/intents/tree/master/example_agent/language)
    to get more insight on the format and naming conventions of language files.
"""
from intents.language_codes import LanguageCode, LANGUAGE_CODES, ensure_language_code
from intents.language.agent_language import agent_language_folder, agent_supported_languages, match_agent_language
from intents.language.intent_language import intent_language_data, IntentResponseGroup, IntentResponseDict, IntentResponse, TextIntentResponse, ImageIntentResponse, QuickRepliesIntentResponse, CardIntentResponse, CustomPayloadIntentResponse, IntentLanguageData, ExampleUtterance, UtteranceChunk, TextUtteranceChunk, EntityUtteranceChunk
from intents.language.entity_language import entity_language_data, EntityEntry



import re

def camel_to_snake_case(name: str) -> str:
    """
    Source: https://stackoverflow.com/a/1176023

    Args:
        name: A CamelCase name

    Returns:
        A snake_case version of the input name
    """
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub('__([A-Z])', r'_\1', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
    return name.lower()



"""
Snips NLU is an open source Python/Rust library for Natural Language Understanding.
It provides intent classification and entity (slot) tagging, and it runs **locally**
as a standard Python library.

:class:`SnipsConnector` is a Connector that doesn't require any other service
(nor remote or local) to run. It makes predictions internally, by calling the
underlying `snips-nlu` library functions.

To use :class:`SnipsConnector` it is necessary to install its optional
dependency group:

.. code-block:: sh

    pip install intents[snips]

More details about Snips can be found at

* https://github.com/snipsco/snips-nlu
* https://snips-nlu.readthedocs.io/
"""
import os
import json
import shutil
import logging
from typing import Union, Type

import snips_nlu

from intents import Intent, Agent, LanguageCode
from intents.language import agent_supported_languages, ensure_language_code
from intents.connectors.interface import Connector, ServiceEntityMappings, FulfillmentRequest
from intents.connectors._experimental.snips.prediction import SnipsPrediction, SnipsPredictionComponent
from intents.connectors._experimental.snips import entities, prediction_format

logger = logging.getLogger(__name__)

class SnipsConnector(Connector):
    """
    This is a :class:`~intents.connectors.interface.Connector` that runs entirely locally, without needing any resident
    service to operate. Predictions are made by calling the `snips-nlu` Python
    API.

    .. warning::

        `SnipsConnector` is **experimental**: expect running into relevant rough
        edges when using it. Its main limitations include:

        * Intent relations are not implemented
        * Prompts for required parameters are not implemented: an exception will be thrown if all the required parameters are not present 
        * :class:`Sys.Email`, :class:`Sys.PhoneNumber` and :class:`Url` entities are patched with empty placeholders
        * :class:`Sys.Date` is only available in English

    Args:
        agent_cls: The Agent class to train the system
        default_session: A default session identifier. Will be generated
            randomly if None
        default_language: Default language for predictions. English will be used
            if None.
    """

    entity_mappings: ServiceEntityMappings = entities.ENTITY_MAPPINGS

    nlu_engine: snips_nlu.SnipsNLUEngine = None
    prediction_component: SnipsPredictionComponent

    def __init__(self,
        agent_cls: Type[Agent],
        default_session: str=None,
        default_language: Union[LanguageCode, str]=None
    ):
        super().__init__(agent_cls, default_session, default_language)
        self.nlu_engines = {
            lang: snips_nlu.SnipsNLUEngine() for lang in agent_supported_languages(agent_cls)
        }
        self.prediction_component = SnipsPredictionComponent(agent_cls, self.entity_mappings)

    def export(self, destination: str):
        """
        Export Agent in the given folder:

        .. code-block:: python

            from example_agent.agent import ExampleAgent
            from intents.connectors._experimental.snips import SnipsConnector

            snips = SnipsConnector(ExampleAgent)
            snips.export("./TMP_SNIPS")

        The export will generate one JSON file per language, they can be loaded
        into Snips as a JSON Dataset.

        Args:
            destination: A folder that will contain exported JSON files
        """
        from intents.connectors._experimental.snips import export

        rendered = export.render(self)

        if os.path.isdir(destination):
            logger.warning("Removing existing export folder: %s", destination)
            shutil.rmtree(destination)
        os.makedirs(destination)

        for lang, data in rendered.items():
            with open(os.path.join(destination, f"agent.{lang.value}.json"), "w") as f:
                json.dump(data, f, indent=4)

    def upload(self):
        """
        As Snips runs locally as a Python library, there is no external service
        to upload the model to. Instead, `upload` will train Snips local models.

        Currently there is no persistence for trained models. This means that
        `upload` should be called every time :class:`SnipsConnector` is instantiated.
        """
        from intents.connectors._experimental.snips import export
        for lang, rendered in export.render(self).items():
            self.nlu_engines[lang].fit(rendered)

    def predict(self, message: str, session: str=None, language: Union[LanguageCode, str]=None) -> SnipsPrediction:
        """
        Predict the given User message in the given session using the given
        language. When `session` or `language` are None, `predict` will use the
        default values that are specified in :meth:`__init__`.

        *predict* will return an instance of :class:`Prediction`, representing
        the service response.

        >>> from intents.connectors._experimental.snips import SnipsConnector
        >>> from example_agent import ExampleAgent
        >>> snips = SnipsConnector(ExampleAgent)
        >>> snips.upload() # This trains the models
        >>> prediction = snips.predict("Hi, my name is Guido")
        >>> prediction.intent
        UserNameGive(user_name='Guido')
        >>> prediction.intent.user_name
        "Guido"
        >>> prediction.fulfillment_text
        "Hi Guido, I'm Bot"
        >>> prediction.confidence
        0.62

        Note that the Italian version of :class:`~example_agent.ExampleAgent`
        won't be trained, as :class:`Sys.Date` is not available for the Italian
        language in Snips.

        Args:
            message: The User message to predict
            session: Any string identifying a conversation
            language: A LanguageCode object, or a ISO 639-1 string (e.g. "en")
        """
        if not language:
            language = self.default_language
        language = ensure_language_code(language)
        parse_result_dict = self.nlu_engines[language].parse(message)
        parse_result = prediction_format.from_dict(parse_result_dict)
        prediction = self.prediction_component.prediction_from_parse_result(parse_result, language)
        return self.prediction_component.fulfill_local(prediction, language)

    def trigger(self, intent: Intent, session: str=None, language: Union[LanguageCode, str]=None) -> SnipsPrediction:
        """
        As Snips runs locally and intent relation resolution is not supported,
        Triggers don't do much more at the moment than returing the intent as it
        was passed in input.

        >>> from intents.connectors._experimental.snips import SnipsConnector
        >>> from example_agent import ExampleAgent, smalltalk
        >>> snips = SnipsConnector(ExampleAgent)
        >>> snips.upload() # This trains the models
        >>> prediction = snips.trigger(smalltalk.AgentNameGive(agent_name='Alice'))
        >>> prediction.intent
        AgentNameGive(agent_name='Alice')
        >>> prediction.fulfillment_text
        "Howdy Human, I'm Alice"
        >>> prediction.confidence
        1.0

        Args:
            intent: The Intent instance to trigger
            session: Any string identifying a conversation
            language: A LanguageCode object, or a ISO 639-1 string (e.g. "en")
        """
        if not language:
            language = self.default_language
        language = ensure_language_code(language)
        prediction = self.prediction_component.prediction_from_intent(intent, language)
        return self.prediction_component.fulfill_local(prediction, language)

    def fulfill(self, fulfillment_request: FulfillmentRequest) -> dict:
        """
        *Not implemented*
        """
        raise NotImplementedError()
